"""Multislice beam propagation simulation. Generate a profile of
the beam that propagates through a bulky sample.
:class:`pyrost.multislice.MLL` generates a transmission profile of
a wedged Multilayer Laue lens (MLL). :class:`pyrost.multislice.MSPropagator`
calculates the beam propagation and spits out the wavefield profile of
the propagated beam together with the transmission profile of the sample
at each slice.

Examples:

    Initialize a MLL object with the :class:`pyrost.multislice.MSParams`
    parameters object `params` with :func:`pyrost.multislice.MLL.import_params`.
    Then you can initialize a multislice beam propagator
    :class:`pyrost.multislice.MSPropagator` with `params` and `mll` and perform
    the multislice beam propagation as follows:

    >>> import pyrost.multislice as ms_sim
    >>> params = ms_sim.MSParams.import_default()
    >>> mll = ms_sim.MLL.import_params(params)
    >>> ms_prgt = ms_sim.MSPropagator(params, mll)
    >>> ms_prgt.beam_propagate() # doctest: +SKIP

    All the results are saved into `ms_prgt.beam_profile` and
    `ms_prgt.smp_profile` attributes.
"""
from __future__ import annotations
from multiprocessing import cpu_count
from typing import Iterable, Optional, Tuple
from tqdm.auto import tqdm
import numpy as np
from .ms_parameters import MSParams
from ..data_container import DataContainer, dict_to_object
from ..bin import mll_profile, FFTW, empty_aligned, rsc_wp

class MLL(DataContainer):
    """
    Multilayer Laue lens class.

    Attributes:
        layers : MLL's bilayers x coordinates [um].
        mat1_r: Fresnel transmission coefficients for the first material,
            that the MLL's bilayers are composed of. The coefficients are the
            ratio of a wavefield propagating through a single slice of the
            lens.
        mat2_r: Fresnel transmission coefficients for the second material,
            that the MLL's bilayers are composed of. The coefficients are the
            ratio of a wavefield propagating through a single slice of the
            lens.
        sigma : Bilayer's interdiffusion length [um].
    """
    attr_set = {'layers', 'mat1_r', 'mat2_r', 'sigma'}
    en_to_wl = 1.239841929761768 # h * c / e [eV * um]

    layers  : np.ndarray
    mat1_r  : complex
    mat2_r  : complex
    sigma   : float

    def __init__(self, mat1_r: complex, mat2_r: complex, sigma: float,
                 layers: np.ndarray) -> None:
        """
        Args:
            mat1_r: Fresnel transmission coefficients for the first material,
                that the MLL's bilayers are composed of. The coefficients are the
                ratio of a wavefield propagating through a single slice of the
                lens.
            mat2_r: Fresnel transmission coefficients for the second material,
                that the MLL's bilayers are composed of. The coefficients are the
                ratio of a wavefield propagating through a single slice of the
                lens.
            sigma : Bilayer's interdiffusion length [um].
            layers : MLL's bilayers x coordinates [um].
        """
        super(MLL, self).__init__(mat1_r=mat1_r, mat2_r=mat2_r, sigma=sigma,
                                  layers=layers)

    @classmethod
    def import_params(cls, params: MSParams) -> MLL:
        """Return a new :class:`MLL` object import from the :class:`MSParams`
        multislice parameters object.

        Args:
            params : Experimental parameters of the multislice beam
                propagation simulation.

        Returns:
            New :class:`MLL` object.
        """
        n_arr = np.arange(params.n_min, params.n_max)
        z_coords = params.z_step * np.arange(params.mll_depth // params.z_step)
        layers = np.sqrt(n_arr * params.focus * params.mll_wl + \
                         0.25 * n_arr**2 * params.mll_wl**2)
        layers = layers * (1.0 - z_coords[:, None] / (2.0 * params.focus))
        mat1_r = params.get_mat1_r(cls.en_to_wl / params.wl)
        mat2_r = params.get_mat2_r(cls.en_to_wl / params.wl)
        sigma = params.mll_sigma
        return cls(mat1_r=mat1_r, mat2_r=mat2_r, sigma=sigma,
                   layers=layers)

    @property
    def n_slices(self) -> int:
        "Total number of slices"
        return self.layers.shape[0]

    @dict_to_object
    def update_interdiffusion(self, sigma: float) -> MLL:
        """Return a new :class:`MLL` object with the updated `sigma`.

        Args:
            sigma : Bilayer's interdiffusion length [um].

        Returns:
            New :class:`MLL` object with the updated `sigma`.
        """
        return {'sigma': sigma}

    @dict_to_object
    def update_materials(self, mat1_r: complex, mat2_r: complex) -> MLL:
        """Return a new :class:`MLL` object with the updated materials `mat1_r`
        and `mat2_r`.

        Args:
            mat1_r, mat2_r : Fresnel transmission coefficients for each of the
                materials, that the MLL's bilayers are composed of. The coefficients
                are the ratio of a wavefield propagating through a single slice of
                the lens.

        Returns:
            New :class:`MLL` object with the updated `mat1_r` and `mat2_r`.
        """
        return {'mat1_r': mat1_r, 'mat2_r': mat2_r}

    def get_span(self) -> Tuple[float, float]:
        """Return the pair of bounds (x_min, x_max) of the MLL
        along the x axis.

        Returns:
            MLL's bounds along the x axis [um].
        """
        return (self.layers.min(), self.layers.max())

    def get_profile(self, x_arr: np.ndarray, output: np.ndarray,
                    num_threads: int=1) -> Iterable:
        """Return a generator, that yields transmission profiles of
        the lens at each slice and writes to `output` array.

        Args:
            x_arr : Coordinates array [um].
            output : Output array.
            num_threads : Number of threads.

        Returns:
            The generator, which yields lens' transmission profiles.
        """
        for idx, layer in enumerate(self.layers):
            output[idx, :] = mll_profile(x_arr=x_arr, layers=layer, t0=self.mat1_r,
                                         t1=self.mat2_r, sigma=self.sigma,
                                         num_threads=num_threads)
            yield output[idx]

class MSPropagator(DataContainer):
    """One-dimensional Multislice beam propagation class.
    Generates beam profile, that propagates through the sample
    using multislice approach.

    Attributes:
        attr_set : Set of attributes in the container which are necessary
            to initialize in the constructor.
        init_set : Set of optional data attributes.

    Notes:
        **Necessary attributes**:

        * sample : Sample's transmission profile generator.
        * num_threads : Number of threads used in the calculations.
        * params : Experimental parameters.

        **Optional attributes**:

        * beam_profile : Beam profiles at each slice.
        * fx_arr : Spatial frequencies array [um^-1].
        * kernel : Diffraction kernel.
        * smp_profile : Sample's transmission profile at each slice.
        * x_arr : Coordinates array in the transverse plane [um].
        * wf_inc : Wavefront at the entry surface.
        * z_arr : Coordinates array along the propagation axis [um].
    """
    attr_set = {'params', 'sample'}
    init_set = {'beam_profile', 'fx_arr', 'kernel', 'num_threads', 'smp_profile', 'x_arr',
                'wf_inc', 'z_arr'}

    # Necessary attributes
    params          : MSParams
    sample          : MLL

    # Automatically generated attributes
    num_threads     : int
    fx_arr          : np.ndarray
    kernel          : np.ndarray
    wf_inc          : np.ndarray
    x_arr           : np.ndarray
    z_arr           : np.ndarray

    # Optional attributes
    beam_profile    : Optional[np.ndarray]
    smp_profile     : Optional[np.ndarray]

    def __init__(self, params: MSParams, sample: MLL, num_threads: Optional[int]=None,
                 **kwargs: np.ndarray) -> None:
        """
        Args:
            params : Experimental parameters.
            sample : Sample class, that generates the sample's
                transmission profile.
            num_threads : Number of threads used in the calculations.
            kwargs : Attributes specified in `init_set`.
        """
        super(MSPropagator, self).__init__(params=params, sample=sample,
                                           num_threads=num_threads, **kwargs)

        self._init_functions(num_threads=lambda: np.clip(1, 64, cpu_count()),
                             x_arr=self.params.get_xcoords, z_arr=self.params.get_zcoords,
                             fx_arr=lambda: np.fft.fftfreq(self.size, self.params.x_step),
                             kernel=lambda: self.params.get_kernel(self.fx_arr) / self.fx_arr.size,
                             wf_inc=self._inc_wavefront)

        self._init_attributes()

    def _inc_wavefront(self):
        wf_inc = np.ones(self.x_arr.shape, dtype=np.complex128)
        x_min, x_max = self.sample.get_span()
        wf_inc[(self.x_arr < x_min) | (self.x_arr > x_max)] = 0.0
        return wf_inc

    @property
    def size(self) -> int:
        "Number of points in a single slice."
        return self.x_arr.size

    @dict_to_object
    def update_inc_wavefront(self, wf_inc: np.ndarray) -> MSPropagator:
        """Return a new :class:`MSPropagator` object with the updated `wf_inc`.

        Args:
            wf_inc : Wavefront at the entry surface.

        Returns:
            A new :class:`MSPropagator` object with the updated `wf_inc`.
        """
        if wf_inc.shape != self.x_arr.shape:
            raise ValueError(f'Wavefront shape must be equal to {self.x_arr.shape:s}')
        if wf_inc.dtype != np.complex128:
            raise ValueError("Wavefront datatype must be 'complex128'")
        return {'beam_profile': None, 'wf_inc': wf_inc}

    def generate_sample(self, verbose: bool=True) -> None:
        """Generate the transmission profile of the sample. The results are
        written to `smp_profile` attribute.

        Args:
            verbose : Set verbosity of the computation process.
        """
        self.smp_profile = empty_aligned((self.sample.n_slices, self.x_arr.size),
                                         dtype='complex128')
        itor = self.sample.get_profile(x_arr=self.x_arr, output=self.smp_profile,
                                       num_threads=self.num_threads)

        if verbose:
            itor = tqdm(enumerate(itor), total=self.sample.n_slices,
                        bar_format='{desc} {percentage:3.0f}% {bar} Slice {n_fmt} / {total_fmt} '\
                                '[{elapsed}<{remaining}, {rate_fmt}{postfix}]')
        for idx, _ in itor:
            if verbose:
                itor.set_description(f"z = {self.z_arr[idx]:.2f} um")

    def beam_propagate(self, verbose: bool=True) -> None:
        """Perform the multislice beam propagation. The results are
        written to `beam_profile` attribute. Generates `smp_profile`
        attribute if it wasn't initialized before.

        Args:
            verbose : Set verbosity of the computation process.
        """
        self.beam_profile = empty_aligned((self.sample.n_slices + 1, self.x_arr.size),
                                          dtype='complex128')
        current_slice = empty_aligned(self.x_arr.size, dtype='complex128')
        self.beam_profile[0] = self.wf_inc
        current_slice[:] = self.wf_inc

        fft_obj = FFTW(current_slice, current_slice, flags=('FFTW_ESTIMATE',),
                       threads=self.num_threads)
        ifft_obj = FFTW(current_slice, self.beam_profile[1], direction='FFTW_BACKWARD',
                        flags=('FFTW_ESTIMATE',), threads=self.num_threads)

        if self.smp_profile is None:
            self.smp_profile = empty_aligned((self.sample.n_slices, self.x_arr.size),
                                             dtype='complex128')
            itor = self.sample.get_profile(x_arr=self.x_arr, output=self.smp_profile,
                                           num_threads=self.num_threads)
        else:
            itor = (layer for layer in self.smp_profile)

        miniters = max(self.sample.n_slices // 100, 1)
        itor = tqdm(itor, total=self.sample.n_slices, miniters=miniters, disable=not verbose,
                    bar_format='{desc} {percentage:3.0f}% {bar} Slice {n_fmt} / {total_fmt} '\
                               '[{elapsed}<{remaining}, {rate_fmt}{postfix}]')

        for idx, layer in enumerate(itor):
            if verbose and not idx % miniters:
                itor.set_description(f"z = {self.z_arr[idx]:.2f} um")
            current_slice *= layer
            fft_obj.execute()
            current_slice *= self.kernel
            ifft_obj.update_arrays(current_slice, self.beam_profile[idx + 1])
            ifft_obj.execute()
            current_slice[:] = ifft_obj.output_array

    def beam_downstream(self, z_arr: np.ndarray, step: Optional[float]=None,
                        return_coords: bool=True, verbose: bool=True,
                        backend: str='fftw') -> Tuple[np.ndarray, np.ndarray]:
        """Return the wavefront at distance `z_arr` downstream from the exit
        surface using Rayleigh-Sommerfeld convolution.

        Args:
            z_arr : Array of distances [um].
            step : Sampling interval of the downstream coordinate array [um].
                Equals to the sampling interval at the exit surface if it's
                None.
            return_coords : Return the coordinates array of the downstream
                plane.
            verbose : Set verbosity of the computation process.
            backend : Choose between numpy ('numpy') or FFTW ('fftw') library
                for the FFT implementation.

        Raises:
            AttributeError : If `beam_profile` has not been generated. Call
                :func:`MSPropagator.beam_propagate` to generate.

        Returns:
            A tuple of two elements ('wavefronts', 'x_arr'). The elements
            are the following:

            * `wavefronts` : Set of wavefronts calculated at the distance
              `z_arr` downstream from the exit surface.
            * `x_arr` : Array of coordinates at the plane downstream [um].
              Only if `return_coords` is True.
        """
        if self.beam_profile is None:
            raise AttributeError('The beam profile has not been generated')
        if step is None:
            step = self.params.x_step
        size = int(self.x_arr.size**2 * max(self.params.x_step, step) * \
                   (step + self.params.x_step) / self.params.wl / np.min(z_arr))
        wf0 = self.beam_profile[-1]
        if size > self.x_arr.size:
            wf0 = np.pad(wf0, ((size - self.x_arr.size) // 2, (size - self.x_arr.size) // 2))
        else:
            size = self.x_arr.size

        itor = np.atleast_1d(z_arr)
        if verbose:
            itor = tqdm(itor, total=itor.size)

        wavefronts = []
        for dist in itor:
            if verbose:
                itor.set_description(f"z = {dist:.2f} um")
            wavefronts.append(rsc_wp(wft=wf0, dx0=self.params.x_step, dx=step,
                                     z=dist, wl=self.params.wl, backend=backend,
                                     num_threads=self.num_threads))
        if return_coords:
            x_arr = step * np.arange(-size // 2, size - size // 2) + np.mean(self.x_arr)
            return np.squeeze(np.stack(wavefronts, axis=1)), x_arr
        return np.squeeze(np.stack(wavefronts, axis=1))



"""Routines for model regression based on nonlinear
least-squares algorithm. :class:`pyrost.AberrationsFit` fit the
lens' aberrations profile with the polynomial function
using nonlinear least-squares algorithm.

Examples:
    Generate a :class:`pyrost.AberrationsFit` object from
    :class:`pyrost.STData` container object `data` as follows:

    .. code-block:: python

        >>> fit_obj = data.get_fit()
        >>> fit = fit_obj.fit(max_order=3)
        >>> print(fit)
        {'c_3': -0.04718488324311934, 'c_4':  0.,
        'fit': array([-9.03305155e-04,  2.14699128e+00, -1.17287983e+03]),
        'ph_fit': array([-9.81298119e-07,  3.49854945e-03, -3.82244504e+00,  1.26179239e+03]),
        'rel_err': array([0.02331385, 0.01966198, 0.01679612]),
        'r_sq': 0.9923840802879347}
"""
from __future__ import annotations
from weakref import ReferenceType
from typing import Any, Dict, Iterable, Optional, Tuple, Union
import numpy as np
from scipy.optimize import least_squares
from .data_container import DataContainer, dict_to_object

class LeastSquares:
    """Basic nonlinear least-squares fit class. Based on
    :func:`scipy.optimize.least_squares`.

    See Also:
        :func:`scipy.optimize.least_squares` : Full nonlinear least-squares
        algorithm description.
    """
    @classmethod
    def model(cls, fit: np.ndarray, x: np.ndarray, roi: Iterable) -> np.ndarray:
        """Return values of polynomial model function.

        Args:
            fit : Array of fit coefficients.
            x : Array of x coordinates.
            roi : Region of interest in the detector plane.

        Returns:
            Array of polynomial function values.
        """
        return np.polyval(fit, x[roi[0]:roi[1]])

    @classmethod
    def errors(cls, fit: np.ndarray, x: np.ndarray, y: np.ndarray,
               roi: Iterable) -> np.ndarray:
        """Return an array of model residuals.

        Args:
            fit : Array of fit coefficients.
            x : Array of x coordinates.
            y : Array of y coordinates.
            roi : Region of interest in the detector plane.

        Returns:
            Array of model residuals.
        """
        return cls.model(fit, x, roi) - y[roi[0]:roi[1]]

    @classmethod
    def fit(cls, x: np.ndarray, y: np.ndarray, max_order: int=2,
            xtol: float=1e-14, ftol: float=1e-14, loss: str='cauchy',
            roi: Optional[Iterable]=None) -> Tuple[np.ndarray, np.ndarray, float]:
        """Fit `x`, `y` with polynomial function using
        :func:`scipy.optimise.least_squares`.

        Args:
            x : Array of x coordinates.
            y : Array of y coordinates.
            max_order : Maximum order of the polynomial model function.
            xtol : Tolerance for termination by the change of the independent variables.
            ftol : Tolerance for termination by the change of the cost function.
            loss : Determines the loss function. The following keyword values are
                allowed:

                * `linear`: ``rho(z) = z``. Gives a standard
                  least-squares problem.
                * `soft_l1` : ``rho(z) = 2 * ((1 + z)**0.5 - 1)``. The smooth
                  approximation of l1 (absolute value) loss. Usually a good
                  choice for robust least squares.
                * `huber` : ``rho(z) = z if z <= 1 else 2*z**0.5 - 1``. Works
                  similarly to 'soft_l1'.
                * `cauchy` (default) : ``rho(z) = ln(1 + z)``. Severely weakens
                  outliers influence, but may cause difficulties in optimization
                  process.
                * `arctan` : ``rho(z) = arctan(z)``. Limits a maximum loss on
                  a single residual, has properties similar to 'cauchy'.

        Returns:
            A tuple of three elements ('fit', 'err', 'r_sq'). The elements are the
            following:

            * `fit` : Array of fit coefficients.
            * `err` : Vector of errors of the `fit` fit coefficients.
            * `r_sq` : ``R**2`` goodness of fit.
        """
        if roi is None:
            roi = (0, x.size)
        fit = least_squares(cls.errors, np.zeros(max_order + 1),
                            loss=loss, args=(x, y, roi), xtol=xtol, ftol=ftol)
        r_sq = 1 - np.sum(cls.errors(fit.x, x, y, roi)**2) / \
               np.sum((y[roi[0]:roi[1]].mean() - y[roi[0]:roi[1]])**2)
        if np.linalg.det(fit.jac.T.dot(fit.jac)):
            cov = np.linalg.inv(fit.jac.T.dot(fit.jac))
            err = np.sqrt(np.sum(fit.fun**2) / (fit.fun.size - fit.x.size) * np.abs(np.diag(cov)))
        else:
            err = 0
        return fit.x, err, r_sq

class AberrationsFit(DataContainer):
    """Least squares optimizer for the lens aberrations' profiles.
    :class:`AberrationsFit` is capable of fitting lens' pixel
    aberrations, deviation  angles, and phase profile with polynomial
    function. Based on :func:`scipy.optimise.least_squares`.

    Attributes:
        kwargs : Necessary and optional attributes specified in the notes
            section.
        det_ap : Aperture angle of a single pixel in detector plane.
        ref_ap : Aperture angle of a single pixel in reference plane.

    Notes:
        **Necessary attributes**:

        * defocus : Defocus distance [m].
        * distance : Sample-to-detector distance [m].
        * pixels : Pixel coordinates [pixels].
        * pixel_aberrations : Pixel aberrations profile [pixels].
        * pixel_size : Pixel's size [m].
        * wavelength : Incoming beam's wavelength [m].

        **Optional attributes**:

        * roi : Region of interest.
        * thetas : Scattering angles [rad].
        * theta_ab : angular displacement profile [rad].
        * phase : aberrations phase profile [rad].

    See Also:
        :func:`scipy.optimize.least_squares` : Full nonlinear least-squares
        algorithm description.
    """
    attr_set = {'defocus', 'distance', 'parent', 'pixels', 'pixel_aberrations',
                'pixel_size', 'wavelength'}
    init_set = {'det_ap', 'phase', 'ref_ap', 'roi', 'thetas', 'theta_ab'}
    x_lookup = {'defocus': 'defocus_x', 'pixel_size': 'x_pixel_size'}
    y_lookup = {'defocus': 'defocus_y', 'pixel_size': 'y_pixel_size'}

    # Necessary attributes
    parent              : ReferenceType
    defocus             : float
    distance            : float
    pixels              : np.ndarray
    pixel_aberrations   : np.ndarray
    pixel_size          : float
    wavelength          : float

    # Automatically generated attributes
    det_ap              : float
    phase               : np.ndarray
    ref_ap              : float
    roi                 : np.ndarray
    thetas              : np.ndarray
    theta_ab            : np.ndarray

    def __init__(self, parent: ReferenceType, **kwargs: Union[float, np.ndarray]) -> None:
        """
        Args:
            parent : The Speckle tracking data container, from which the object
                is derived.
            kwargs : Necessary and optional attributes specified in the notes
                section.

        Raises:
            ValueError : If any of the necessary attributes has not been provided.
        """
        super(AberrationsFit, self).__init__(parent=parent, **kwargs)

        self._init_functions(det_ap=lambda: self.pixel_size / self.distance,
                             ref_ap=lambda: np.abs(self.det_ap * self.defocus / self.distance),
                             roi=lambda: np.array([0, self.pixels.size]),
                             thetas=lambda: self.pixels * self.det_ap,
                             theta_ab=lambda: self.pixel_aberrations * self.ref_ap,
                             phase=lambda: self.wnumber * np.cumsum(self.theta_ab * self.pixel_size))

        self._init_attributes()
        self.phase -= self.phase.mean()

    @property
    def wnumber(self) -> float:
        return 2.0 * np.pi / self.wavelength

    @dict_to_object
    def crop_data(self, roi: Iterable) -> AberrationsFit:
        """Return a new :class:`AberrationsFit` object with the updated `roi`.

        Args:
            roi : Region of interest in the detector plane.

        Returns:
            New :class:`AberrationsFit` object with the updated `roi`.
        """
        return {'roi': np.asarray(roi, dtype=int)}

    @dict_to_object
    def remove_linear_term(self, fit: Optional[np.ndarray]=None, xtol: float=1e-14,
                           ftol: float=1e-14, loss: str='cauchy') -> AberrationsFit:
        """Return a new :class:`AberrationsFit` object with the linear term
        removed from `pixel_aberrations` profile.

        Args:
            fit : Fit coefficients of a first order polynomial. Inferred from
                `pixel_aberrations` by fitting a line if None.
            xtol : Tolerance for termination by the change of the independent
                variables.
            ftol : Tolerance for termination by the change of the cost function.
            loss : Determines the loss function. The following keyword values
                are allowed:

                * `linear` : ``rho(z) = z``. Gives a standard
                  least-squares problem.
                * `soft_l1` : ``rho(z) = 2 * ((1 + z)**0.5 - 1)``. The smooth
                  approximation of l1 (absolute value) loss. Usually a good
                  choice for robust least squares.
                * `huber` : ``rho(z) = z if z <= 1 else 2*z**0.5 - 1``. Works
                  similarly to 'soft_l1'.
                * `cauchy` (default) : ``rho(z) = ln(1 + z)``. Severely weakens
                  outliers influence, but may cause difficulties in optimization
                  process.
                * `arctan` : ``rho(z) = arctan(z)``. Limits a maximum loss on
                  a single residual, has properties similar to 'cauchy'.

        Returns:
            New :class:`AberrationsFit` object with the updated `pixel_aberrations`
            and `phase`.
        """
        if fit is None:
            fit = LeastSquares.fit(x=self.pixels, y=self.pixel_aberrations,
                                   roi=self.roi, max_order=1, xtol=xtol,
                                   ftol=ftol, loss=loss)[0]
        pixel_aberrations = self.pixel_aberrations - self.model(fit)
        return {'pixel_aberrations': pixel_aberrations, 'theta_ab': None, 'phase': None}

    @dict_to_object
    def update_center(self, center: float) -> AberrationsFit:
        """Return a new :class:`AberrationsFit` object with the pixels
        centered around `center`.

        Args:
            center : Index of the zerro scattering angle or direct
                beam pixel.

        Returns:
            New :class:`AberrationsFit` object with the updated `pixels`,
            `phase`, and `pixel_aberrations`.
        """
        if center <= self.pixels[0]:
            pixels = self.pixels - center
            return {'pixels': pixels}
        elif center >= self.pixels[-1]:
            pixels = center - self.pixels
            idxs = np.argsort(self.pixels)
            return {'pixels': pixels[idxs], 'thetas': None,
                    'pixel_aberrations': -self.pixel_aberrations[idxs], 'theta_ab': None}
        else:
            raise ValueError('Origin must be outside of the region of interest')

    @dict_to_object
    def update_phase(self) -> AberrationsFit:
        """Return a new :class:`AberrationsFit` object with the updated `phase`.

        Returns:
            New :class:`AberrationsFit` object with the updated `phase`.
        """
        return {'theta_ab': None, 'phase': None}

    def get(self, attr: str, value: Optional[Any]=None) -> Any:
        """Return a dataset with `mask` and `roi` applied. Return `value`
        if the attribute is not found.

        Args:
            attr : Attribute to fetch.
            value : Return if `attr` is not found.

        Returns:
            `attr` dataset with `mask` and `roi` applied. `value` if
            `attr` is not found.
        """
        if attr in self:
            val = super(AberrationsFit, self).get(attr)
            if not val is None:
                if attr in ['phase', 'pixels', 'pixel_aberrations', 'theta_ab']:
                    val = val[self.roi[0]:self.roi[1]]
            return val

        return value

    def model(self, fit: np.ndarray) -> np.ndarray:
        """Return the polynomial function values of lens' deviation angles fit.

        Args:
            fit : Lens` pixel aberrations fit coefficients.

        Returns:
            Array of polynomial function values.
        """
        return LeastSquares.model(fit, self.pixels, [0, self.pixels.size])

    def pix_to_phase(self, fit: np.ndarray) -> np.ndarray:
        """Convert fit coefficients from pixel aberrations fit to aberrations
        phase fit.

        Args:
            fit : Lens' pixel aberrations fit coefficients.

        Returns:
            Lens` phase aberrations fit coefficients.
        """
        nfit = np.zeros(fit.size + 1)
        nfit[:-1] = self.wnumber * fit * self.ref_ap * self.pixel_size
        nfit[:-1] /= np.arange(1, fit.size + 1)[::-1]
        nfit[-1] = -self.model(nfit).mean()
        return nfit

    def phase_to_pix(self, ph_fit: np.ndarray) -> np.ndarray:
        """Convert fit coefficients from pixel aberrations fit to aberrations
        phase fit.

        Args:
            ph_fit : Lens` phase aberrations fit coefficients.

        Returns:
            Lens' pixel aberrations fit coefficients.
        """
        fit = ph_fit[:-1] * self.wavelength / (2.0 * np.pi * self.ref_ap * self.pixel_size)
        fit *= np.arange(1, ph_fit.size)[::-1]
        return fit

    def fit(self, max_order: int=2, xtol: float=1e-14, ftol: float=1e-14,
            loss: str='cauchy') -> Dict[str, Union[float, np.ndarray]]:
        """Fit lens' pixel aberrations with polynomial function using
        :func:`scipy.optimise.least_squares`.

        Args:
            max_order : Maximum order of the polynomial model function.
            xtol : Tolerance for termination by the change of the independent
                variables.
            ftol : Tolerance for termination by the change of the cost function.
            loss : Determines the loss function. The following keyword values
                are allowed:

                * `linear` : ``rho(z) = z``. Gives a standard
                  least-squares problem.
                * `soft_l1` : ``rho(z) = 2 * ((1 + z)**0.5 - 1)``. The smooth
                  approximation of l1 (absolute value) loss. Usually a good
                  choice for robust least squares.
                * `huber` : ``rho(z) = z if z <= 1 else 2*z**0.5 - 1``. Works
                  similarly to 'soft_l1'.
                * `cauchy` (default) : ``rho(z) = ln(1 + z)``. Severely weakens
                  outliers influence, but may cause difficulties in optimization
                  process.
                * `arctan` : ``rho(z) = arctan(z)``. Limits a maximum loss on
                  a single residual, has properties similar to 'cauchy'.

        Returns:
            A dictionary with the fitting information. The following elements are
            defined inside:

            * `c_3` : Third order aberrations coefficient [rad / mrad^3].
            * `c_4` : Fourth order aberrations coefficient [rad / mrad^4].
            * `fit` : Array of the polynomial function coefficients of the
              pixel aberrations fit.
            * `ph_fit` : Array of the polynomial function coefficients of
              the phase aberrations fit.
            * `rel_err` : Vector of relative errors of the fit coefficients.
            * `r_sq` : ``R**2`` goodness of fit.

        See Also:
            :func:`scipy.optimize.least_squares` : Full nonlinear least-squares
            algorithm description.
        """
        fit, err, r_sq = LeastSquares.fit(x=self.pixels, y=self.pixel_aberrations,
                                          roi=self.roi, max_order=max_order,
                                          xtol=xtol, ftol=ftol, loss=loss)
        ph_fit = self.pix_to_phase(fit)
        c_3, c_4 = 0., 0.

        if ph_fit.size >= 4:
            c_3 = ph_fit[-4] * (self.distance / self.pixel_size)**3 * 1e-9

        if ph_fit.size >= 5:
            c_4 = ph_fit[-5] * (self.distance / self.pixel_size)**4 * 1e-12

        return {'c_3': c_3, 'c_4': c_4, 'fit': fit, 'ph_fit': ph_fit,
                'rel_err': np.abs(err / fit), 'r_sq': r_sq}

    def fit_phase(self, max_order: int=3, xtol: float=1e-14, ftol: float=1e-14,
                  loss: str='linear') -> Dict[str, Union[float, np.ndarray]]:
        """Fit lens' phase aberrations with polynomial function using
        :func:`scipy.optimise.least_squares`.

        Args:
            max_order : Maximum order of the polynomial model function.
            xtol : Tolerance for termination by the change of the independent
                variables.
            ftol : Tolerance for termination by the change of the cost function.
            loss : Determines the loss function. The following keyword values
                are allowed:

                * `linear` : ``rho(z) = z``. Gives a standard
                  least-squares problem.
                * `soft_l1` : ``rho(z) = 2 * ((1 + z)**0.5 - 1)``. The smooth
                  approximation of l1 (absolute value) loss. Usually a good
                  choice for robust least squares.
                * `huber` : ``rho(z) = z if z <= 1 else 2*z**0.5 - 1``. Works
                  similarly to 'soft_l1'.
                * `cauchy` (default) : ``rho(z) = ln(1 + z)``. Severely weakens
                  outliers influence, but may cause difficulties in optimization
                  process.
                * `arctan` : ``rho(z) = arctan(z)``. Limits a maximum loss on
                  a single residual, has properties similar to 'cauchy'.

        Returns:
            A dictionary with the fitting information. The following elements are
            defined inside:

            * `c_3` : Third order aberrations coefficient [rad / mrad^3].
            * `c_4` : Fourth order aberrations coefficient [rad / mrad^4].
            * `fit` : Array of the polynomial function coefficients of the
              pixel aberrations fit.
            * `ph_fit` : Array of the polynomial function coefficients of
              the phase aberrations fit.
            * `rel_err` : Vector of relative errors of the fit coefficients.
            * `r_sq` : ``R**2`` goodness of fit.

        See Also:
            :func:`scipy.optimize.least_squares` : Full nonlinear least-squares
            algorithm description.
        """
        ph_fit, err, r_sq = LeastSquares.fit(x=self.pixels, y=self.phase, roi=self.roi,
                                             max_order=max_order, xtol=xtol, ftol=ftol,
                                             loss=loss)
        fit = self.phase_to_pix(ph_fit)
        c_3, c_4 = 0., 0.

        if ph_fit.size >= 4:
            c_3 = ph_fit[-4] * (self.distance / self.pixel_size)**3 * 1e-9

        if ph_fit.size >= 5:
            c_4 = ph_fit[-5] * (self.distance / self.pixel_size)**4 * 1e-12

        return {'c_3': c_3, 'c_4': c_4, 'fit': fit, 'ph_fit': ph_fit,
                'rel_err': np.abs(err / ph_fit)[:-1], 'r_sq': r_sq}



"""
Examples:

    :func:`pyrost.simulation.STParams.import_default` generates the speckle tracking
    experimental parameters, which could be later parsed to :class:`pyrost.simulation.STSim`
    in order to perform the simulation.

    >>> import pyrost.simulation as st_sim
    >>> st_params = st_sim.STParams.import_default()
    >>> print(st_params)
    {'exp_geom': {'defocus': 100.0, 'det_dist': 2000000.0, 'n_frames': 300, '...': '...'},
     'detector': {'detx_size': 2000, 'dety_size': 1000, 'pix_size': 55.0},
     'source': {'p0': 200000.0, 'th_s': 0.0002, 'wl': 7.29e-05}, '...': '...'}
"""
from __future__ import annotations
import os
from typing import Dict, Iterable, Iterator, Tuple, Union, Optional
from multiprocessing import cpu_count
import numpy as np
from ..ini_parser import INIParser, ROOT_PATH
from ..bin import bar_positions, barcode_profile, gaussian_kernel, gaussian_filter

ST_PARAMETERS = os.path.join(ROOT_PATH, 'config/st_parameters.ini')

class STParams(INIParser):
    """Container with the simulation parameters for the wavefront propagation.

    Attributes:
        kwargs : Experimental parameters enlisted in :ref:`st-parameters`.

    See Also:
        :ref:`st-parameters` : Full list of experimental parameters.
    """
    attr_dict = {'exp_geom': ('defocus', 'det_dist', 'n_frames',
                              'step_size', 'step_rnd'),
                 'detector': ('detx_size', 'dety_size', 'pix_size'),
                 'source':   ('p0', 'th_s', 'wl'),
                 'lens':     ('alpha', 'ap_x', 'ap_y', 'focus', 'ab_cnt'),
                 'barcode':  ('bar_atn', 'bar_rnd', 'bar_sigma', 'bar_size',
                              'bulk_atn', 'offset'),
                 'system':   ('num_threads', 'seed')}

    fmt_dict = {'exp_geom': 'float', 'exp_geom/n_frames': 'int',
                'detector': 'int', 'detector/pix_size': 'float',
                'source': 'float', 'lens': 'float', 'barcode': 'float',
                'system': 'int'}

    # exp_geom attributes
    defocus     : float
    det_dist    : float
    n_frames    : int
    step_size   : float
    step_rnd    : float

    # detector attributes
    detx_size   : int
    dety_size   : int
    pix_size    : float

    # source attributes
    p0          : float
    th_s        : float
    wl          : float

    # lens attributes
    alpha       : float
    ap_x        : float
    ap_y        : float
    focus       : float
    ab_cnt      : float

    # barcode attributes
    bar_atn     : float
    bar_rnd     : float
    bar_sigma   : float
    bar_size    : float
    bulk_atn    : float
    offset      : float

    # system attributes
    num_threads : int
    seed        : int

    def __init__(self, barcode: Dict[str, float], detector: Dict[str, Union[int, float]],
                 exp_geom: Dict[str, Union[int, float]], lens: Dict[str, float],
                 source: Dict[str, float], system: Dict[str, int]) -> None:
        """
        Args:
            barcode : A dictionary of barcode sample parameters. The following elements
                are accepted:

                * `bar_size` : Average bar's size [um].
                * `bar_sigma` : Bar bluriness width [um].
                * `bar_atn` : Bar's attenuation coefficient [0.0 - 1.0].
                * `bulk_atn` : Barcode's bulk attenuation coefficient [0.0 - 1.0].
                * `bar_rnd` : Bar's coordinates random deviation [0.0 - 1.0].
                * `offset` : Barcode's offset at the beginning and at the end
                  of the scan from the detector's bounds [um].

            detector : A dictionary of detector parameters. The following elements are 
                accepted:

                * `detx_size` : Detector's size along the horizontal axis in pixels.
                * `dety_size` : Detector's size along the vertical axis in pixels.
                * `pix_size` : Detector's pixel size [um].

            exp_geom : A dictionary of experimental geometry parameters. The following elements
                are accepted:

                * `defocus` : Lens' defocus distance [um].
                * `det_dist` : Distance between the barcode and the detector [um].
                * `step_size` : Scan step size [um].
                * `n_frames` : Number of frames.

            lens : A dictionary of lens parameters. The following elements are accepted:

                * `ap_x` : Lens' aperture size along the x axis [um].
                * `ap_y` : Lens' aperture size along the y axis [um].
                * `focus` : Focal distance [um].
                * `alpha` : Third order aberrations coefficient [rad / mrad^3].
                * `ab_cnt` : Lens' aberrations center point [0.0 - 1.0].

            source : A dictionary of X-ray source parameters. The following elements are
                accepted:

                * `p0` : Source beam flux [cnt / s].
                * `wl` : Source beam's wavelength [um].
                * `th_s` : Source rocking curve width [rad].

            system : A dictionary of calculation parameters. The following elements are
                accepted:

                * `seed` : Seed used in all the pseudo-random number generations.
                * `num_threads` : Number of threads used in the calculations.
        """
        super(STParams, self).__init__(barcode=barcode, detector=detector,
                                       exp_geom=exp_geom, lens=lens, source=source,
                                       system=system)
        if self.seed <= 0:
            self.update_seed()
        if self.num_threads <= 0:
            self.update_threads()

    @classmethod
    def _lookup_dict(cls) -> Dict[str, str]:
        lookup = {}
        for section in cls.attr_dict:
            for option in cls.attr_dict[section]:
                lookup[option] = section
        return lookup

    def __iter__(self) -> Iterator[str]:
        return self._lookup.__iter__()

    def __contains__(self, attr: str) -> bool:
        return attr in self._lookup

    def __repr__(self) -> str:
        return self._format(self.export_dict()).__repr__()

    def __str__(self) -> str:
        return self._format(self.export_dict()).__str__()

    def keys(self) -> Iterable[str]:
        return list(self)

    @classmethod
    def import_default(cls, **kwargs: Union[int, float]) -> STParams:
        """Return the default :class:`STParams`. Extra arguments
        override the default values if provided.

        Args:
            kwargs : Simulation parameters enlisted in :ref:`st-parameters`.

        Returns:
            An :class:`STParams` object with the default parameters.
        """
        return cls.import_ini(ST_PARAMETERS, **kwargs)

    @classmethod
    def import_ini(cls, ini_file: str, **kwargs: Union[int, float]) -> STParams:
        """Initialize a :class:`STParams` object with an
        ini file.

        Args:
            ini_file : Path to the ini file.
            kwargs : Experimental parameters enlisted in :ref:`st-parameters`.
                Override the parameters imported from the `ini_file`.

        Returns:
            A :class:`STParams` object with all the attributes imported
            from the ini file.
        """
        attr_dict = cls._import_ini(ini_file)
        for option, section in cls._lookup_dict().items():
            if option in kwargs:
                attr_dict[section][option] = kwargs[option]
        return cls(**attr_dict)

    def update_seed(self, seed: Optional[int]=None) -> None:
        """Update seed used in pseudo-random number generation.

        Args:
            seed : New seed value. Chosen randomly if None.
        """
        if seed is None or seed <= 0:
            seed = np.random.default_rng().integers(0, np.iinfo(np.int_).max, endpoint=False)
        self.seed = seed

    def update_threads(self, num_threads: Optional[int]=None) -> None:
        """Update number of threads used in calculcations.

        Args:
            num_threads : Number of threads used in the computations.
        """
        if num_threads is None or num_threads <= 0 or num_threads > 64:
            num_threads = np.clip(1, 64, cpu_count())
        self.num_threads = num_threads

    def x_wavefront_size(self) -> int:
        r"""Return wavefront array size along the x axis, that
        satisfies the sampling condition.

        Returns:
            Array size.

        Notes:
            The sampling condition when we propagate the wavefield from
            the lens plane to the sample plane is as follows:

            .. math::
                N_x >= \frac{\mathrm{max} \left( \Delta x_{lens}, \Delta x_{sample} \right)
                (\Delta x_{lens} + \Delta x_{sample})}{\lambda d} =
                \frac{4 a_x^2 \mathrm{max} \left( f, df \right)}{f^2 \lambda}

            Whereupon, the sampling condition when we propagate the wavefield to the detector
            plane is:

            .. math::
                N_x >= \frac{\Delta x_{sample} \Delta x_{det}}{\lambda d} =
                \frac{2 a_x n_{x} \Delta_{pix} df}{f \lambda d_{det}}
        """
        nx_ltos = int(4 * self.ap_x**2 * max(self.focus, np.abs(self.defocus)) \
                      / self.focus**2 / self.wl)
        nx_stod = int(2 * self.detx_size * self.pix_size * self.ap_x * np.abs(self.defocus) \
                      / self.focus / self.wl / self.det_dist)
        return max(nx_ltos, nx_stod)

    def y_wavefront_size(self) -> int:
        r"""Return wavefront array size along the y axis, that
        satisfies the sampling condition.

        Returns:
            Array size.

        Notes:
            The sampling condition when we propagate the wavefield from
            the lens plane to the sample plane is as follows:

            .. math::
                N_y >= \frac{\mathrm{max} \left( \Delta y_{lens}, \Delta y_{sample} \right)
                (\Delta y_{lens} + \Delta y_{sample})}{\lambda d} =
                \frac{8 a_y^2}{(f + df) \lambda}

            Whereupon, the sampling condition when we propagate the wavefield to the detector
            plane is:

            .. math::
                N_y >= \frac{\Delta y_{sample} \Delta y_{det}}{\lambda d} =
                \frac{2 a_y n_{y} \Delta_{pix}}{\lambda d_{det}}
        """
        ny_ltos = int(8 * self.ap_y**2 / (self.focus + self.defocus) / self.wl)
        ny_stod = int(2 * self.dety_size * self.pix_size * self.ap_y / self.wl / self.det_dist)
        return max(ny_ltos, ny_stod)

    def lens_x_wavefront(self, n_x: Optional[int]=None,
                         return_step: bool=False) -> Tuple[np.ndarray, float]:
        r"""Return wavefields at the lens plane along x axis.

        Args:
            n_x : Array size along the x axis. Equals to
                :func:`STParams.x_wavefront_size` if it's None.
            return_step : Return the step size along x axis if it's True.

        Returns:
            A tuple of two elements ('u0_x', 'dx'). The elements
            are the following:

            * `u0_x` : Wavefront along the x axis.
            * `dx` : Step size along the x axis [um]. Only if `return_step` is True.

        Notes:
            The exit-surface at the lens plane:

            .. math::
                U_0(x) = \Pi(a_x x) \exp
                \left[ -\frac{j \pi x^2}{\lambda f} + j \alpha
                \left( \frac{x - x_{ab_cnt}}{f} \right)^3 \right]
        """
        if n_x is None:
            n_x = self.x_wavefront_size()

        dx = 2.0 * self.ap_x / n_x
        x0 = dx * np.arange(-n_x // 2, n_x // 2)
        x_cnt = (self.ab_cnt - 0.5) * self.ap_x
        u0_x = np.exp(1j * np.pi * x0 * x0 / self.wl / self.focus - \
                      1e9j * self.alpha * ((x0 - x_cnt) / self.focus)**3)
        u0_x[np.abs(x0) > 0.5 * self.ap_x] = 0

        if return_step:
            return u0_x, dx
        return u0_x

    def lens_y_wavefront(self, n_y: Optional[int]=None,
                         return_step: bool=False) -> Tuple[np.ndarray, float]:
        r"""Return wavefields at the lens plane along y axis.

        Args:
            n_y : Array size along the y axis. Equals to
                :func:`STParams.y_wavefront_size` if it's None.
            return_step : Return the step size along y axis if it's True.

        Returns:
            A tuple of two elements ('u0_y', 'dy'). The elements
            are the following:

            * `u0_y` : Wavefront along the y axis.
            * `dy` : Step size along the y axis [um]. Only if `return_step` is True.

        Notes:
            The exit-surface at the lens plane:

            .. math::
                U_0(y) = \Pi(a_y y)
        """
        if n_y is None:
            n_y = self.y_wavefront_size()

        dy = 2.0 * self.ap_y / n_y
        y0 = dy * np.arange(-n_y // 2, n_y // 2)
        u0_y = np.ones(n_y, dtype=np.complex128)
        u0_y[np.abs(y0) > 0.5 * self.ap_y] = 0

        if return_step:
            return u0_y, dy
        return u0_y

    def beam_span(self, dist: float) -> Tuple[float, float]:
        """Return beam span along the x axis at distance `dist`
        from the focal plane.

        Args:
            dist : Distance from the focal plane [um].

        Returns:
            Tuple of two items ('th_lb', 'th_ub'). The elements are
            the following:

            * `th_lb` : Beam's lower bound [um].
            * `th_ub` : Beam's upper bound [um].
        """
        th_lb = -0.5 * self.ap_x / self.focus + self.wl / np.pi * self.alpha * \
                3.75e8 * (self.ap_x / self.focus)**2 / dist
        th_ub = 0.5 * self.ap_x / self.focus + self.wl / np.pi * self.alpha * \
                3.75e8 * (self.ap_x / self.focus)**2 / dist
        return np.tan(th_lb) * dist, np.tan(th_ub) * dist

    def bar_positions(self, dist: float, rnd_dev: bool=True) -> np.ndarray:
        """Generate a coordinate array of barcode's bar positions at distance
        `dist` from focal plane.

        Args:
            dist : Distance from the focal plane [um].
            rnd_dev : Randomize positions if it's True.

        Returns:
            Array of barcode's bar coordinates.

        See Also:
            :func:`pyrost.bin.bar_positions` : Full details of randomized barcode
            steps generation algorithm.
        """
        x0, x1 = self.beam_span(dist)
        seed = self.seed if rnd_dev else -1
        return bar_positions(x0=x0 + self.offset, b_dx=self.bar_size, rd=self.bar_rnd,
                             x1=x1 + self.step_size * self.n_frames - self.offset,
                             seed=seed)

    def sample_positions(self) -> np.ndarray:
        """Generate an array of sample's translations with random deviation.

        Returns:
            Array of sample translations [um].
        """
        rng = np.random.default_rng(self.seed)
        rnd_arr = 2 * self.step_rnd * (rng.random(self.n_frames) - 0.5)
        return self.step_size * (np.arange(self.n_frames) + rnd_arr)

    def barcode_profile(self, x_arr: np.ndarray, dx: float, bars: np.ndarray) -> np.ndarray:
        """Generate a barcode's transmission profile at `x_arr` coordinates.

        Args:
            x_arr : Array of the coordinates, where the transmission
                coefficients are calculated [um].
            dx : Sampling interval of the coordinate array [um].
            bars : Array of barcode's bar positions [um].

        Returns:
            Barcode's transmission profile.

        See Also:
            :func:`pyrost.bin.barcode_profile` : Full details of barcode's
            transmission profile generation algorithm.
        """
        return barcode_profile(x_arr=x_arr, bars=bars, bulk_atn=self.bulk_atn,
                               bar_atn=self.bar_atn, bar_sigma=self.bar_sigma,
                               num_threads=self.num_threads)

    def source_curve(self, dist: float, step: float) -> np.ndarray:
        """Return source's rocking curve profile at `dist` distance from
        the lens.

        Args:
            dist : Distance from the lens to the rocking curve profile [um].
            step : Sampling interval [um].

        Returns:
            Source's rocking curve profile.
        """
        return gaussian_kernel(sigma=dist * self.th_s / step)



import os
import sys
from setuptools import setup, find_packages
from distutils.core import Extension
import numpy

try:
    from Cython.Build import cythonize
except ImportError:
    USE_CYTHON = False
else:
    USE_CYTHON = True

ext = '.pyx' if USE_CYTHON else '.c'
extension_args = {'language': 'c',
                  'extra_compile_args': ['-fopenmp', '-std=c99'],
                  'extra_link_args': ['-lgomp', '-Wl,-rpath,/usr/local/lib'],
                  'libraries': ['gsl', 'gslcblas', 'fftw3', 'fftw3f', 'fftw3_omp', 'fftw3f_omp'],
                  'library_dirs': ['/usr/local/lib',
                                   os.path.join(sys.prefix, 'lib')],
                  'include_dirs': [numpy.get_include(),
                                   os.path.join(sys.prefix, 'include'),
                                   os.path.join(os.path.dirname(__file__), 'pyrost/include'),
                                   os.path.join(os.path.dirname(__file__), 'pyrost/bin')]}

src_files = ["pyrost/include/pocket_fft.c", "pyrost/include/fft_functions.c",
             "pyrost/include/array.c", "pyrost/include/routines.c", "pyrost/include/median.c"]
extensions = [Extension(name='pyrost.bin.simulation',
                        sources=['pyrost/bin/simulation' + ext,] + src_files, **extension_args),
              Extension(name='pyrost.bin.pyfftw',
                        sources=['pyrost/bin/pyfftw' + ext,], **extension_args),
              Extension(name='pyrost.bin.pyrost',
                        sources=['pyrost/bin/pyrost' + ext,], **extension_args),]

if USE_CYTHON:
    extensions = cythonize(extensions, annotate=True, language_level="3",
                           compiler_directives={'cdivision': True,
                                                'boundscheck': False,
                                                'wraparound': False,
                                                'binding': True,
                                                'embedsignature': False})

with open('README.md', 'r') as readme:
    long_description = readme.read()

setup(name='pyrost',
      version='0.7.5',
      author='Nikolay Ivanov',
      author_email="nikolay.ivanov@desy.de",
      long_description=long_description,
      long_description_content_type='text/markdown',
      url="https://github.com/simply-nicky/pyrost",
      packages=find_packages(),
      include_package_data=True,
      package_data={'pyrost': ['config/*.ini', 'ini_templates/*.ini', 'data/*.npz']},
      install_requires=['h5py', 'numpy', 'scipy'],
      extras_require={'interactive': ['matplotlib', 'jupyter', 'pyximport', 'ipykernel', 'ipywidgets']},
      ext_modules=extensions,
      classifiers=[
          "Programming Language :: Python",
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
          "Operating System :: OS Independent"
      ],
      python_requires='>=3.6')



"""Log protocol (:class:`pyrost.LogProtocol`) is a helper class to
retrieve the data from the log files generated at the Sigray laboratory,
which contain the readouts from the motors and other instruments
during the speckle tracking scan.

Examples:
    Generate the default built-in log protocol:

    >>> import pyrost as rst
    >>> rst.LogProtocol()
    {'log_keys': {'det_dist': ['Session logged attributes', 'Z-LENSE-DOWN_det_dist'],
    'exposure': ['Type: Method', 'Exposure'], 'n_steps': ['Type: Scan', 'Points count'],
    '...': '...'}, 'datatypes': {'det_dist': 'float', 'exposure': 'float', 'n_steps':
    'int', '...': '...'}}

    Generate the default log file converter:

    >>> rst.KamzikConverter()
    {'fs_vec': array([-5.5e-05,  0.0e+00,  0.0e+00]), 'protocol': {'datatypes':
    {'lens_down_dist': 'float', 'lens_up_dist': 'float', 'exposure': 'float', '...': '...'},
    'log_keys': {'lens_down_dist': ['Z-LENSE-DOWN_det_dist'], 'lens_up_dist':
    ['Z-LENSE-UP_det_dist'], 'exposure': ['Exposure'], '...': '...'}, 'part_keys':
    {'lens_down_dist': 'Session logged attributes', 'lens_up_dist': 'Session logged attributes',
    'exposure': 'Type: Method', '...': '...'}}, 'ss_vec': array([ 0.0e+00, -5.5e-05,  0.0e+00])}
"""
from __future__ import annotations
import os
import re
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union
import numpy as np
from .data_container import DataContainer, dict_to_object
from .ini_parser import ROOT_PATH, INIParser

LOG_PROTOCOL = os.path.join(ROOT_PATH, 'config/log_protocol.ini')

class LogProtocol(INIParser):
    """Log file protocol class. Contains log file keys to retrieve
    and the data types of the corresponding values.

    Attributes:
        datatypes : Dictionary with attributes' datatypes. 'float', 'int',
            'bool', or 'str' are allowed.
        log_keys : Dictionary with attributes' log file keys.
        part_keys : Dictionary with the part names inside the log file
            where the attributes are stored.
    """
    attr_dict   = {'datatypes': ('ALL',), 'log_keys': ('ALL',), 'part_keys': ('ALL',)}
    fmt_dict    = {'datatypes': 'str', 'log_keys': 'str', 'part_keys': 'str'}
    unit_dict   = {'percent': 1e-2, 'mm,mdeg': 1e-3, 'µm,um,udeg,µdeg': 1e-6,
                   'nm,ndeg': 1e-9, 'pm,pdeg': 1e-12}

    datatypes   : Dict[str, str]
    log_keys    : Dict[str, List[str]]
    part_keys   : Dict[str, str]

    def __init__(self, datatypes: Dict[str, str], log_keys: Dict[str, List[str]],
                 part_keys: Dict[str, str]) -> None:
        """
        Args:
            datatypes : Dictionary with attributes' datatypes. 'float', 'int',
                'bool', or 'str' are allowed.
            log_keys : Dictionary with attributes' log file keys.
            part_keys : Dictionary with the part names inside the log file
                where the attributes are stored.
        """
        log_keys = {attr: val for attr, val in log_keys.items() if attr in datatypes}
        datatypes = {attr: val for attr, val in datatypes.items() if attr in log_keys}
        super(LogProtocol, self).__init__(datatypes=datatypes, log_keys=log_keys,
                                          part_keys=part_keys)

    @classmethod
    def import_default(cls, datatypes: Optional[Dict[str, str]]=None,
                       log_keys: Optional[Dict[str, List[str]]]=None,
                       part_keys: Optional[Dict[str, str]]=None) -> LogProtocol:
        """Return the default :class:`LogProtocol` object. Extra arguments
        override the default values if provided.

        Args:
            datatypes : Dictionary with attributes' datatypes. 'float', 'int',
                or 'bool' are allowed.
            log_keys : Dictionary with attributes' log file keys.
            part_keys : Dictionary with the part names inside the log file
                where the attributes are stored.

        Returns:
            A :class:`LogProtocol` object with the default parameters.
        """
        return cls.import_ini(LOG_PROTOCOL, datatypes, log_keys, part_keys)

    @classmethod
    def import_ini(cls, ini_file: str, datatypes: Optional[Dict[str, str]]=None,
                   log_keys: Optional[Dict[str, List[str]]]=None,
                   part_keys: Optional[Dict[str, str]]=None) -> LogProtocol:
        """Initialize a :class:`LogProtocol` object class with an
        ini file.

        Args:
            ini_file : Path to the ini file. Load the default log protocol if None.
            datatypes : Dictionary with attributes' datatypes. 'float', 'int',
                or 'bool' are allowed. Initialized with `ini_file` if None.
            log_keys : Dictionary with attributes' log file keys. Initialized with
                `ini_file` if None.
            part_keys : Dictionary with the part names inside the log file
                where the attributes are stored. Initialized with `ini_file`
                if None.

        Returns:
            A :class:`LogProtocol` object with all the attributes imported
            from the ini file.
        """
        kwargs = cls._import_ini(ini_file)
        if not datatypes is None:
            kwargs['datatypes'].update(**datatypes)
        if not log_keys is None:
            kwargs['log_keys'].update(**log_keys)
        if not part_keys is None:
            kwargs['part_keys'].update(**part_keys)
        return cls(datatypes=kwargs['datatypes'], log_keys=kwargs['log_keys'],
                   part_keys=kwargs['part_keys'])

    @classmethod
    def _get_unit(cls, key: str) -> float:
        for unit_key in cls.unit_dict:
            units = unit_key.split(',')
            for unit in units:
                if unit in key:
                    return cls.unit_dict[unit_key]
        return 1.0

    @classmethod
    def _has_unit(cls, key: str) -> bool:
        has_unit = False
        for unit_key in cls.unit_dict:
            units = unit_key.split(',')
            for unit in units:
                has_unit |= (unit in key)
        return has_unit

    def load_attributes(self, path: str) -> Dict[str, Any]:
        """Return attributes' values from a log file at
        the given `path`.

        Args:
            path : Path to the log file.

        Returns:
            Dictionary with the attributes retrieved from
            the log file.
        """
        if not isinstance(path, str):
            raise ValueError('path must be a string')
        with open(path, 'r') as log_file:
            log_str = ''
            for line in log_file:
                if line.startswith('# '):
                    log_str += line.strip('# ')
                else:
                    break

        # List all the sector names
        part_keys = list(self.part_keys.values())

        # Divide log into sectors
        parts_list = [part for part in re.split('(' + '|'.join(part_keys) + \
                      '|--------------------------------)\n*', log_str) if part]

        # Rearange sectors into a dictionary
        parts = {}
        for idx, part in enumerate(parts_list):
            if part in part_keys:
                if part == 'Session logged attributes':
                    attr_keys, attr_vals = parts_list[idx + 1].strip('\n').split('\n')
                    parts['Session logged attributes'] = ''
                    for key, val in zip(attr_keys.split(';'), attr_vals.split(';')):
                        parts['Session logged attributes'] += key + ': ' + val + '\n'
                else:
                    val = parts_list[idx + 1]
                    match = re.search(r'Device:.*\n', val)
                    if match:
                        name = match[0].split(': ')[-1][:-1]
                        parts[part + ', ' + name] = val

        # Populate attributes dictionary
        attr_dict = {part_name: {} for part_name in parts}
        for part_name, part in parts.items():
            for attr, part_key in self.part_keys.items():
                if part_key in part_name:
                    for log_key in self.log_keys[attr]:
                        # Find the attribute's mention and divide it into a key and value pair
                        match = re.search(log_key + r'.*\n', part)
                        if match:
                            raw_str = match[0]
                            raw_val = raw_str.strip('\n').split(': ')[1]
                            # Extract a number string
                            val_num = re.search(r'[-]*\d+[.]*\d*', raw_val)
                            dtype = self.known_types[self.datatypes[attr]]
                            attr_dict[part_name][attr] = dtype(val_num[0] if val_num else raw_val)
                            # Apply unit conversion if needed
                            if np.issubdtype(dtype, np.floating):
                                attr_dict[part_name][attr] *= self._get_unit(raw_str)
        return attr_dict

    def load_data(self, path: str, idxs: Optional[Iterable[int]]=None,
                  return_idxs=False) -> Tuple[Dict[str, np.ndarray], np.ndarray]:
        """Retrieve the main data array from the log file.

        Args:
            path : Path to the log file.
            idxs : Array of data indices to load. Loads info for all
                the frames if None.
            return_idxs : Return a set of indices loaded from the log file.

        Returns:
            Dictionary with data fields and their names retrieved from the
            log file.
        """
        if idxs is not None:
            idxs = np.asarray(idxs)
            idxs.sort()

        line_count = 0
        with open(path, 'r') as log_file:
            for line_idx, line in enumerate(log_file):
                if line.startswith('# '):
                    if 'WARNING' not in line:
                        keys_line = line.strip('# ')
                else:
                    data_line = line

                    if idxs is None:
                        skiprows = line_idx
                        max_rows = None
                        break

                    if idxs.size == 0:
                        skiprows = line_idx
                        max_rows = 0
                        break

                    if line_count == idxs[0]:
                        skiprows = line_idx
                    if line_count == idxs[-1]:
                        max_rows = line_idx - skiprows + 1
                        break

                    line_count += 1

        keys = keys_line.strip('\n').split(';')
        data_strings = data_line.strip('\n').split(';')

        dtypes = {'names': [], 'formats': []}
        converters = {}
        for idx, (key, val) in enumerate(zip(keys, data_strings)):
            dtypes['names'].append(key)
            unit = self._get_unit(key)
            if 'float' in key:
                dtypes['formats'].append(np.dtype(float))
                converters[idx] = lambda item, unit=unit: unit * float(item)
            elif 'int' in key:
                if self._has_unit(key):
                    converters[idx] = lambda item, unit=unit: unit * float(item)
                    dtypes['formats'].append(np.dtype(float))
                else:
                    dtypes['formats'].append(np.dtype(int))
            elif 'Array' in key:
                dtypes['formats'].append(np.ndarray)
                func = lambda part, unit=unit: unit * float(part)
                conv = lambda item, func=func: np.asarray(list(map(func, item.strip(b' []').split(b','))))
                converters[idx] = conv
            else:
                dtypes['formats'].append('<S' + str(len(val)))
                converters[idx] = lambda item: item.strip(b' []')

        txt_dict = {}
        txt_tuple = np.loadtxt(path, delimiter=';', converters=converters,
                               dtype=dtypes, unpack=True, skiprows=skiprows,
                               max_rows=max_rows)

        if idxs is None:
            txt_dict.update(zip(keys, txt_tuple))
            idxs = np.arange(txt_tuple[0].size)
        elif idxs.size == 0:
            txt_dict.update(zip(keys, txt_tuple))
        else:
            txt_dict.update({key: np.atleast_1d(data)[idxs - np.min(idxs)]
                             for key, data in zip(keys, txt_tuple)})

        if return_idxs:
            return txt_dict, idxs
        return txt_dict

class KamzikConverter(DataContainer):
    """A converter class, that generates CXI datasets aceeptable by :class:`pyrost.STData`
    from Kamzik log files.

    Attributes:
        protocol : Log file protocol.
        fs_vec : Fast (horizontal) scan detector axis.
        ss_vec : Slow (vertical) scan detector axis.
        idxs : Frame indices read from a log file.
        log_attr : Dictionary of log attributes.
        log_data : Dictionary of log datasets.
    """
    attr_set = {'protocol', 'fs_vec', 'ss_vec'}
    init_set = {'idxs', 'log_attr', 'log_data'}

    cxi_attrs = {'basis_vectors': 'basis_vectors', 'dist_down': 'distance', 'dist_up': 'distance',
                 'sim_translations': 'translations', 'log_translations': 'translations',
                 'x_pixel_size': 'x_pixel_size', 'y_pixel_size': 'y_pixel_size'}

    protocol:   LogProtocol
    fs_vec:     np.ndarray
    ss_vec:     np.ndarray

    idxs:       Optional[np.ndarray]
    log_attr:   Optional[Dict[str, Any]]
    log_data:   Optional[Dict[str, Any]]

    def __init__(self, protocol: LogProtocol=LogProtocol.import_default(),
                 fs_vec: np.ndarray=np.array([-55e-6, 0., 0.]),
                 ss_vec: np.ndarray=np.array([0., -55e-6, 0.]),
                 idxs: Optional[np.ndarray]=None, log_attr: Optional[Dict[str, Any]]=None,
                 log_data: Optional[Dict[str, Any]]=None) -> None:
        """
        Args:
            protocol : Log file protocol.
            fs_vec : Fast (horizontal) scan detector axis.
            ss_vec : Slow (vertical) scan detector axis.
            idxs : Frame indices read from a log file.
            log_attr : Dictionary of log attributes read from a log file.
            log_data : Dictionary of log datasets read from a log file.
        """
        super(KamzikConverter, self).__init__(protocol=protocol, fs_vec=fs_vec, ss_vec=ss_vec,
                                              idxs=idxs, log_attr=log_attr, log_data=log_data)

    @property
    def n_frames(self):
        return None if self.idxs is None else self.idxs.size

    @property
    def x_pixel_size(self):
        return np.sqrt((self.fs_vec * self.fs_vec).sum())

    @property
    def y_pixel_size(self):
        return np.sqrt((self.ss_vec * self.ss_vec).sum())

    @dict_to_object
    def read_logs(self, log_path: str, idxs: Optional[Iterable[int]]=None) -> KamzikConverter:
        """Read a log file under the path `log_path`. Read out only the frame indices
        defined by `idxs`. If `idxs` is None, read the whole log file.

        Args:
            log_path : Path to the log file.
            idxs : List of indices to read. Read the whole log file if None.

        Returns:
            A new :class:`KamzikConverter` object with `log_attr`, `log_data`, and `idxs`
            updated.
        """
        log_attr = self.protocol.load_attributes(log_path)
        log_data, idxs = self.protocol.load_data(log_path, idxs=idxs, return_idxs=True)
        return {'log_attr': log_attr, 'log_data': log_data, 'idxs': idxs}

    def find_log_part_key(self, attr: str) -> Optional[str]:
        """Find a name of the log dictionary corresponding to an attribute
        name `attr`.

        Args:
            attr : A name of the attribute to find.

        Returns:
            A name of the log dictionary, corresponding to the given attribute
            name `attr`.
        """
        log_attr = self.get('log_attr', {})
        log_keys = self.protocol.log_keys.get(attr, [])
        for part in log_attr:
            for log_key in log_keys:
                if log_key in part:
                    return part
        return None

    def find_log_attribute(self, attr: str, part_key: Optional[str]=None) -> Optional[Any]:
        """Find a value in the log attributes corresponding to an
        attribute name `attr`.

        Args:
            attr : A name of the attribute to find.
            part_key : Search in the given part of the log dictionary if provided.

        Returns:
            Value of the log attribute. Returns None if nothing is found.
        """
        if part_key is None:
            part_key = self.protocol.part_keys.get(attr, '')
        log_attr = self.get('log_attr', {})
        part_dict = log_attr.get(part_key, {})
        value = part_dict.get(attr, None)
        return value

    def find_log_dataset(self, attr: str) -> Optional[np.ndarray]:
        """Find a dataset in the log data corresponding to an
        attribute name `attr`.

        Args:
            attr : A name of the attribute to find.

        Returns:
            Dataset for the given attribute. Returns None if nothing is found.
        """
        log_keys = self.protocol.log_keys.get(attr, [])
        log_data = self.get('log_data', {})
        for data_key, log_dset in log_data.items():
            for log_key in log_keys:
                if log_key in data_key:
                    return log_dset
        return None

    def _is_basis_vectors(self) -> bool:
        return self.n_frames is not None

    def _is_dist_down(self) -> bool:
        return self.find_log_attribute('lens_down_dist') is not None

    def _is_dist_up(self) -> bool:
        return self.find_log_attribute('lens_up_dist') is not None

    def _is_log_translations(self) -> bool:
        return (self.find_log_attribute('x_sample') is not None and
                self.find_log_attribute('y_sample') is not None and
                self.find_log_attribute('z_sample') is not None and
                (self.find_log_dataset('x_sample') is not None or
                 self.find_log_dataset('y_sample') is not None or
                 self.find_log_dataset('z_sample') is not None))

    def _is_sim_translations(self) -> bool:
        return (self.find_log_attribute('x_sample') is not None and
                self.find_log_attribute('y_sample') is not None and
                self.find_log_attribute('z_sample') is not None and
                (self.find_log_part_key('x_sample') is not None or
                 self.find_log_part_key('y_sample') is not None or
                 self.find_log_part_key('z_sample') is not None))

    def cxi_keys(self) -> List[str]:
        """Return a list of available CXI attributes.

        Returns:
            List of available CXI attributes.
        """
        cxi_dict = {'basis_vectors': self._is_basis_vectors,
                    'dist_down': self._is_dist_down,
                    'dist_up': self._is_dist_up,
                    'sim_translations': self._is_sim_translations,
                    'log_translations': self._is_log_translations}
        return [attr for attr, func in cxi_dict.items() if func()]

    def _get_basis_vectors(self) -> np.ndarray:
        return np.stack((np.tile(self.ss_vec, (self.n_frames, 1)),
                         np.tile(self.fs_vec, (self.n_frames, 1))), axis=1)

    def _get_dist_down(self) -> float:
        return self.find_log_attribute('lens_down_dist')

    def _get_dist_up(self) -> float:
        return self.find_log_attribute('lens_up_dist')

    def _get_sim_translations(self) -> np.ndarray:
        translations = np.tile((self.find_log_attribute('x_sample'),
                                self.find_log_attribute('y_sample'),
                                self.find_log_attribute('z_sample')), (self.n_frames, 1))
        translations = np.nan_to_num(translations)

        step_sizes, n_steps = [], []
        for scan_motor, unit_vec in zip(['x_sample', 'y_sample', 'z_sample'],
                                        [np.array([1., 0., 0.]),
                                         np.array([0., 1., 0.]),
                                         np.array([0., 0., 1.])]):
            part_key = self.find_log_part_key(scan_motor)
            if part_key is not None:
                step_sizes.append(self.log_attr[part_key].get('step_size') * unit_vec)
                n_steps.append(self.log_attr[part_key].get('n_points'))

        steps = np.tensordot(np.stack(np.mgrid[[slice(0, n) for n in n_steps]], axis=0),
                             np.stack(step_sizes, axis=0), (0, 0)).reshape(-1, 3)
        return translations + steps

    def _get_log_translations(self) -> np.ndarray:
        translations = np.tile((self.find_log_attribute('x_sample'),
                                self.find_log_attribute('y_sample'),
                                self.find_log_attribute('z_sample')), (self.n_frames, 1))
        translations = np.nan_to_num(translations)

        for idx, scan_motor in enumerate(['x_sample', 'y_sample', 'z_sample']):
            dset = self.find_log_dataset(scan_motor)
            if dset is not None:
                translations[:dset.size, idx] = dset
        return translations

    def cxi_get(self, attrs: Union[str, List[str]]) -> Dict[str, Any]:
        """Convert Kamzik log files data into CXI attributes, that are accepted by
        :class:`pyrost.STData` container. To see full list of available CXI
        attributes, use :func:`KamzikConverter.cxi_keys`.

        Args:
            attrs: List of CXI attributes to generate. The method will raise an error
                if any of the attributes is unavailable.

        Raises:
            ValueError : If any of attributes in `attrs` in unavailable.

        Returns:
            A dictionary of CXI attributes, that are accepted by :class:`pyrost.STData`
            container.
        """
        cxi_dict = {'basis_vectors': self._get_basis_vectors,
                    'dist_down': self._get_dist_down,
                    'dist_up': self._get_dist_up,
                    'sim_translations': self._get_sim_translations,
                    'log_translations': self._get_log_translations}
        data = {'x_pixel_size': self.x_pixel_size, 'y_pixel_size': self.y_pixel_size}
        cxi_keys = self.cxi_keys()
        for attr in attrs:
            if attr in cxi_keys:
                data[self.cxi_attrs[attr]] = cxi_dict[attr]()
            else:
                raise ValueError(f"CXI attribute '{attr}' is unavailable")

        return data



"""`pyrost`_ is a Python library for wavefront metrology and sample imaging
based on ptychographic speckle tracking algorithm. This project
takes over Andrew Morgan's `Ptychographic Speckle Tracking project`_
as an improved version aiming to add robustness to the optimisation
algorithm in the case of the high noise present in the measured data.

.. _pyrost: https://github.com/simply-nicky/pyrost
.. _Ptychographic Speckle Tracking project: https://github.com/andyofmelbourne/speckle-tracking

(c) Nikolay Ivanov, 2020.
"""
from __future__ import absolute_import

from .cxi_protocol import CXIProtocol, CXIStore
from .log_protocol import LogProtocol, KamzikConverter
from .data_processing import Transform, Crop, Downscale, Mirror, ComposeTransforms, STData
from .rst_update import SpeckleTracking
from .bfgs import BFGS
from .aberrations_fit import AberrationsFit
from . import bin



import numpy as np
import cython
from libc.string cimport memcmp
from libc.math cimport sqrt
from libc.stdlib cimport abort, malloc, free

# Numpy must be initialized. When using numpy from C or Cython you must
# *ALWAYS* do that, or you will have segfaults
np.import_array()

# Helper functions

cdef int extend_mode_to_code(str mode) except -1:
    if mode == 'constant':
        return EXTEND_CONSTANT
    elif mode == 'nearest':
        return EXTEND_NEAREST
    elif mode == 'mirror':
        return EXTEND_MIRROR
    elif mode == 'reflect':
        return EXTEND_REFLECT
    elif mode == 'wrap':
        return EXTEND_WRAP
    else:
        raise RuntimeError('boundary mode not supported')

cdef np.ndarray check_array(np.ndarray array, int type_num):
    cdef np.ndarray out
    cdef int tn = np.PyArray_TYPE(array)
    if not np.PyArray_IS_C_CONTIGUOUS(array):
        array = np.PyArray_GETCONTIGUOUS(array)

    if tn != type_num:
        out = np.PyArray_SimpleNew(array.ndim, <np.npy_intp *>array.shape, type_num)
        np.PyArray_CastTo(out, array)
    else:
        out = array

    return out

cdef np.ndarray number_to_array(object num, np.npy_intp rank, int type_num):
    cdef np.npy_intp *dims = [rank,]
    cdef np.ndarray arr = <np.ndarray>np.PyArray_SimpleNew(1, dims, type_num)
    cdef int i
    for i in range(rank):
        arr[i] = num
    return arr

cdef np.ndarray normalize_sequence(object inp, np.npy_intp rank, int type_num):
    # If input is a scalar, create a sequence of length equal to the
    # rank by duplicating the input. If input is a sequence,
    # check if its length is equal to the length of array.
    cdef np.ndarray array, out
    cdef int tn
    if np.PyArray_IsAnyScalar(inp):
        out = number_to_array(inp, rank, type_num)

    elif np.PyArray_Check(inp):
        array = <np.ndarray>inp
        tn = np.PyArray_TYPE(array)
        if tn != type_num:
            out = np.PyArray_SimpleNew(array.ndim, <np.npy_intp *>array.shape, type_num)
            np.PyArray_CastTo(out, array)
        else:
            out = array

    elif isinstance(inp, (list, tuple)):
        out = <np.ndarray>np.PyArray_FROM_OTF(inp, type_num, np.NPY_ARRAY_C_CONTIGUOUS)

    else:
        raise ValueError("Wrong sequence argument type")

    cdef np.npy_intp size = np.PyArray_SIZE(out)
    if size != rank:
        raise ValueError("Sequence argument must have length equal to input rank")
    return out

def next_fast_len(unsigned target, str backend='numpy'):
    r"""Find the next fast size of input data to fft, for zero-padding, etc.
    FFT algorithms gain their speed by a recursive divide and conquer strategy.
    This relies on efficient functions for small prime factors of the input length.
    Thus, the transforms are fastest when using composites of the prime factors handled
    by the fft implementation. If there are efficient functions for all radices <= n,
    then the result will be a number x >= target with only prime factors < n. (Also
    known as n-smooth numbers)

    Args:
        target (int) : Length to start searching from. Must be a positive integer.
        backend (str) : Find n-smooth number for the FFT implementation from the numpy
            ('numpy') or FFTW ('fftw') library.

    Returns:
        int : The smallest fast length greater than or equal to `target`.
    """
    if target < 0:
        raise ValueError('Target length must be positive')
    if backend == 'fftw':
        return next_fast_len_fftw(target)
    elif backend == 'numpy':
        return good_size(target)
    else:
        raise ValueError('{:s} is invalid backend'.format(backend))

def fft_convolve(np.ndarray array not None, np.ndarray kernel not None, int axis=-1,
                 str mode='constant', double cval=0.0, str backend='numpy',
                 unsigned num_threads=1):
    """Convolve a multi-dimensional `array` with one-dimensional `kernel` along the
    `axis` by means of FFT. Output has the same size as `array`.

    Args:
        array (numpy.ndarray) : Input array.
        kernel (numpy.ndarray) : Kernel array.
        axis (int) : Array axis along which convolution is performed.
        mode (str) : The mode parameter determines how the input array is extended
            when the filter overlaps a border. Default value is 'constant'. The
            valid values and their behavior is as follows:

            * `constant`, (k k k k | a b c d | k k k k) : The input is extended by
              filling all values beyond the edge with the same constant value, defined
              by the `cval` parameter.
            * `nearest`, (a a a a | a b c d | d d d d) : The input is extended by
              replicating the last pixel.
            * `mirror`, (c d c b | a b c d | c b a b) : The input is extended by
              reflecting about the center of the last pixel. This mode is also sometimes
              referred to as whole-sample symmetric.
            * `reflect`, (d c b a | a b c d | d c b a) : The input is extended by
              reflecting about the edge of the last pixel. This mode is also sometimes
              referred to as half-sample symmetric.
            * `wrap`, (a b c d | a b c d | a b c d) : The input is extended by wrapping
              around to the opposite edge.

        cval (float) :  Value to fill past edges of input if mode is 'constant'. Default
            is 0.0.
        backend (str) : Choose between numpy ('numpy') or FFTW ('fftw') library for the FFT
            implementation.
        num_threads (int) : Number of threads used in the calculations.

    Returns:
        numpy.ndarray : A multi-dimensional array containing the discrete linear
        convolution of `array` with `kernel`.
    """
    cdef int fail = 0
    cdef int ndim = array.ndim
    axis = axis if axis >= 0 else ndim + axis
    axis = axis if axis <= ndim - 1 else ndim - 1
    cdef np.npy_intp ksize = np.PyArray_DIM(kernel, 0)
    cdef int _mode = extend_mode_to_code(mode)
    cdef np.npy_intp *dims = array.shape
    cdef unsigned long *_dims = <unsigned long *>dims

    cdef int type_num
    if np.PyArray_ISCOMPLEX(array) or np.PyArray_ISCOMPLEX(kernel):
        type_num = np.NPY_COMPLEX128
    else:
        type_num = np.NPY_FLOAT64

    cdef np.ndarray out = <np.ndarray>np.PyArray_SimpleNew(ndim, dims, type_num)
    cdef void *_out = np.PyArray_DATA(out)
    cdef void *_inp
    cdef void *_krn

    if np.PyArray_ISCOMPLEX(array) or np.PyArray_ISCOMPLEX(kernel):
        array = check_array(array, np.NPY_COMPLEX128)
        kernel = check_array(kernel, np.NPY_COMPLEX128)

        _inp = np.PyArray_DATA(array)
        _krn = np.PyArray_DATA(kernel)

        with nogil:
            if backend == 'fftw':
                fail = cfft_convolve_fftw(<double complex *>_out, <double complex *>_inp, ndim,
                                          _dims, <double complex *>_krn, ksize, axis, _mode,
                                          <double complex>cval, num_threads)
            elif backend == 'numpy':
                fail = cfft_convolve_np(<double complex *>_out, <double complex *>_inp, ndim,
                                        _dims, <double complex *>_krn, ksize, axis, _mode,
                                        <double complex>cval, num_threads)
            else:
                raise ValueError('{:s} is invalid backend'.format(backend))
    else:
        array = check_array(array, np.NPY_FLOAT64)
        kernel = check_array(kernel, np.NPY_FLOAT64)

        _inp = np.PyArray_DATA(array)
        _krn = np.PyArray_DATA(kernel)

        with nogil:
            if backend == 'fftw':
                fail = rfft_convolve_fftw(<double *>_out, <double *>_inp, ndim, _dims,
                                          <double *>_krn, ksize, axis, _mode, cval, num_threads)
            elif backend == 'numpy':
                fail = rfft_convolve_np(<double *>_out, <double *>_inp, ndim, _dims,
                                        <double *>_krn, ksize, axis, _mode, cval, num_threads)
            else:
                raise ValueError('{:s} is invalid backend'.format(backend))

    if fail:
        raise RuntimeError('C backend exited with error.')
    return out

def rsc_wp(np.ndarray wft not None, double dx0, double dx, double z,
           double wl, int axis=-1, str backend='numpy', unsigned num_threads=1):
    r"""Wavefront propagator based on Rayleigh-Sommerfeld convolution
    method [RSC]_. Propagates a wavefront `wft` by `z` distance downstream.
    You can choose between 'fftw' and 'numpy' backends for FFT calculations.

    Args: 
        wft (numpy.ndarray) : Initial wavefront.
        dx0 (float) : Sampling interval at the plane upstream [um].
        dx (float) : Sampling interval at the plane downstream [um].
        z (float) : Propagation distance [um].
        wl (float) : Incoming beam's wavelength [um].
        axis (int) : Axis of `wft` array along which the calculation is performed.
        backend (str) : Choose between numpy ('numpy') or FFTW ('fftw') library
            for the FFT  implementation.
        num_threads (int) : Number of threads used in the calculations.

    Returns:
        numpy.ndarray : Propagated wavefront.

    Raises:
        RuntimeError : If 'numpy' backend exits with eror during the calculation.
        ValueError : If `backend` option is invalid.

    Notes:
        The Rayleigh-Sommerfeld diffraction integral transform is defined as:

        .. math::
            u^{\prime}(x^{\prime}) = \frac{z}{j \sqrt{\lambda}} \int_{-\infty}^{+\infty}
            u(x) \mathrm{exp} \left[-j k r(x, x^{\prime}) \right] dx
        
        with

        .. math::
            r(x, x^{\prime}) = \left[ (x - x^{\prime})^2 + z^2 \right]^{1 / 2}

    References:
        .. [RSC] V. Nascov and P. C. Logofătu, "Fast computation algorithm
                for the Rayleigh-Sommerfeld diffraction formula using
                a type of scaled convolution," Appl. Opt. 48, 4310-4319
                (2009).
    """
    wft = check_array(wft, np.NPY_COMPLEX128)

    cdef int fail = 0
    cdef np.npy_intp isize = np.PyArray_SIZE(wft)
    cdef int ndim = wft.ndim
    axis = axis if axis >= 0 else ndim + axis
    axis = axis if axis <= ndim - 1 else ndim - 1
    cdef np.npy_intp *dims = wft.shape
    cdef unsigned long *_dims = <unsigned long *>dims
    cdef np.ndarray out = <np.ndarray>np.PyArray_SimpleNew(ndim, dims, np.NPY_COMPLEX128)
    cdef complex *_out = <complex *>np.PyArray_DATA(out)
    cdef complex *_inp = <complex *>np.PyArray_DATA(wft)
    with nogil:
        if backend == 'fftw':
            fail = rsc_fftw(_out, _inp, ndim, _dims, axis, dx0, dx, z, wl, num_threads)
        elif backend == 'numpy':
            fail = rsc_np(_out, _inp, ndim, _dims, axis, dx0, dx, z, wl, num_threads)
        else:
            raise ValueError('{:s} is invalid backend'.format(backend))
    if fail:
        raise RuntimeError('C backend exited with error.')
    return out

def fraunhofer_wp(np.ndarray wft not None, double dx0, double dx, double z,
                  double wl, int axis=-1, str backend='numpy', unsigned num_threads=1):
    r"""Fraunhofer diffraction propagator. Propagates a wavefront `wft` by `z`
    distance downstream. You can choose between 'fftw' and 'numpy' backends for
    FFT calculations.

    Args: 
        wft (numpy.ndarray) : Initial wavefront.
        dx0 (float) : Sampling interval at the plane upstream [um].
        dx (float) : Sampling interval at the plane downstream [um].
        z (float) : Propagation distance [um].
        wl (float) : Incoming beam's wavelength [um].
        axis (int) : Axis of `wft` array along which the calculation is performed.
        backend (str) : Choose between numpy ('numpy') or FFTW ('fftw') library
            for the FFT  implementation.
        num_threads (int) : Number of threads used in the calculations.

    Returns:
        numpy.ndarray : Propagated wavefront.

    Raises:
        RuntimeError : If 'numpy' backend exits with eror during the calculation.
        ValueError : If `backend` option is invalid.

    Notes:
        The Fraunhofer integral transform is defined as:

        .. math::
            u^{\prime}(x^{\prime}) = \frac{e^{-j k z}}{j \sqrt{\lambda z}}
            e^{-\frac{j k}{2 z} x^{\prime 2}} \int_{-\infty}^{+\infty} u(x)
            e^{j\frac{2 \pi}{\lambda z} x x^{\prime}} dx
    """
    wft = check_array(wft, np.NPY_COMPLEX128)

    cdef int fail = 0
    cdef np.npy_intp isize = np.PyArray_SIZE(wft)
    cdef int ndim = wft.ndim
    axis = axis if axis >= 0 else ndim + axis
    axis = axis if axis <= ndim - 1 else ndim - 1
    cdef np.npy_intp *dims = wft.shape
    cdef unsigned long *_dims = <unsigned long *>dims
    cdef np.ndarray out = <np.ndarray>np.PyArray_SimpleNew(ndim, dims, np.NPY_COMPLEX128)
    cdef complex *_out = <complex *>np.PyArray_DATA(out)
    cdef complex *_inp = <complex *>np.PyArray_DATA(wft)
    with nogil:
        if backend == 'fftw':
            fail = fraunhofer_fftw(_out, _inp, ndim, _dims, axis, dx0, dx, z, wl, num_threads)
        elif backend == 'numpy':
            fail = fraunhofer_np(_out, _inp, ndim, _dims, axis, dx0, dx, z, wl, num_threads)
        else:
            raise ValueError('{:s} is invalid backend'.format(backend))
    if fail:
        raise RuntimeError('C backend exited with error.')
    return out

def gaussian_kernel(double sigma, unsigned order=0, double truncate=4.):
    """Discrete Gaussian kernel.
    
    Args:
        sigma (float) : Standard deviation for Gaussian kernel.
        order (int) : The order of the filter. An order of 0 corresponds to
            convolution with a Gaussian kernel. A positive order corresponds
            to convolution with that derivative of a Gaussian. Default is 0.
        truncate (float) : Truncate the filter at this many standard deviations.
            Default is 4.0.
    
    Returns:
        numpy.ndarray : Gaussian kernel.
    """
    cdef np.npy_intp radius = <np.npy_intp>(sigma * truncate)
    cdef np.npy_intp *dims = [2 * radius + 1,]
    cdef np.ndarray out = <np.ndarray>np.PyArray_SimpleNew(1, dims, np.NPY_FLOAT64)
    cdef double *_out = <double *>np.PyArray_DATA(out)
    with nogil:
        gauss_kernel1d(_out, sigma, order, dims[0], 1)
    return out

def gaussian_filter(np.ndarray inp not None, object sigma not None, object order not None=0,
                    str mode='reflect', double cval=0., double truncate=4., str backend='numpy',
                    unsigned num_threads=1):
    r"""Multidimensional Gaussian filter. The multidimensional filter is implemented as
    a sequence of 1-D FFT convolutions.

    Args:
        inp (numpy.ndarray) : The input array.
        sigma (Union[float, List[float]]): Standard deviation for Gaussian kernel. The standard
            deviations of the Gaussian filter are given for each axis as a sequence, or as a
            single number, in which case it is equal for all axes.
        order (Union[int, List[int]]): The order of the filter along each axis is given as a
            sequence of integers, or as a single number. An order of 0 corresponds to convolution
            with a Gaussian kernel. A positive order corresponds to convolution with that
            derivative of a Gaussian.
        mode (str) : The mode parameter determines how the input array is extended when the
            filter overlaps a border. Default value is 'reflect'. The valid values and their
            behavior is as follows:

            * `constant`, (k k k k | a b c d | k k k k) : The input is extended by filling all
              values beyond the edge with the same constant value, defined by the `cval`
              parameter.
            * `nearest`, (a a a a | a b c d | d d d d) : The input is extended by replicating
              the last pixel.
            * `mirror`, (c d c b | a b c d | c b a b) : The input is extended by reflecting
              about the center of the last pixel. This mode is also sometimes referred to as
              whole-sample symmetric.
            * `reflect`, (d c b a | a b c d | d c b a) : The input is extended by reflecting
              about the edge of the last pixel. This mode is also sometimes referred to as
              half-sample symmetric.
            * `wrap`, (a b c d | a b c d | a b c d) : The input is extended by wrapping around
              to the opposite edge.

        cval (float) : Value to fill past edges of input if mode is 'constant'. Default is 0.0.
        truncate (float) : Truncate the filter at this many standard deviations. Default is 4.0.
        backend (str) : Choose between numpy ('numpy') or FFTW ('fftw') backend library for the
            FFT implementation.
        num_threads (int) : Number of threads.
    
    Returns:
        numpy.ndarray : Returned array of the same shape as `input`.
    """

    cdef int ndim = inp.ndim
    cdef np.ndarray sigmas = normalize_sequence(sigma, ndim, np.NPY_FLOAT64)
    cdef np.ndarray orders = normalize_sequence(order, ndim, np.NPY_UINT32)
    cdef double *_sig = <double *>np.PyArray_DATA(sigmas)
    cdef unsigned *_ord = <unsigned *>np.PyArray_DATA(orders)

    cdef int n
    for n in range(ndim):
        if inp.shape[n] == 1:
            sigmas[n] = 0.0

    cdef int fail = 0
    cdef int _mode = extend_mode_to_code(mode)
    cdef np.npy_intp *dims = inp.shape
    cdef unsigned long *_dims = <unsigned long *>dims

    cdef int type_num
    if np.PyArray_ISCOMPLEX(inp):
        type_num = np.NPY_COMPLEX128
    else:
        type_num = np.NPY_FLOAT64

    cdef np.ndarray out = <np.ndarray>np.PyArray_SimpleNew(ndim, dims, type_num)
    cdef void *_out = np.PyArray_DATA(out)

    cdef void *_inp
    cdef int inp_tn = np.PyArray_TYPE(inp)
    if np.PyArray_ISCOMPLEX(inp):
        inp = check_array(inp, np.NPY_COMPLEX128)
        _inp = <double *>np.PyArray_DATA(inp)

        with nogil:
            if backend == 'fftw':
                fail = gauss_filter_c(<double complex *>_out, <double complex *>_inp,
                                      ndim, _dims, _sig, _ord, _mode, <double complex>cval,
                                      truncate, num_threads, cfft_convolve_fftw)
            elif backend == 'numpy':
                fail = gauss_filter_c(<double complex *>_out, <double complex *>_inp,
                                      ndim, _dims, _sig, _ord, _mode, <double complex>cval,
                                      truncate, num_threads, cfft_convolve_np)
            else:
                raise ValueError('{:s} is invalid backend'.format(backend))

    else:
        inp = check_array(inp, np.NPY_FLOAT64)
        _inp = <double *>np.PyArray_DATA(inp)

        with nogil:
            if backend == 'fftw':
                fail = gauss_filter_r(<double *>_out, <double *>_inp, ndim, _dims, _sig,
                                      _ord, _mode, cval, truncate, num_threads, rfft_convolve_fftw)
            elif backend == 'numpy':
                fail = gauss_filter_r(<double *>_out, <double *>_inp, ndim, _dims, _sig,
                                      _ord, _mode, cval, truncate, num_threads, rfft_convolve_np)
            else:
                raise ValueError('{:s} is invalid backend'.format(backend))

    if fail:
        raise RuntimeError('C backend exited with error.')
    return check_array(out, inp_tn)

def gaussian_gradient_magnitude(np.ndarray inp not None, object sigma not None, str mode='reflect',
                                double cval=0.0, double truncate=4.0, str backend='numpy',
                                unsigned num_threads=1):
    r"""Multidimensional gradient magnitude using Gaussian derivatives. The multidimensional
    filter is implemented as a sequence of 1-D FFT convolutions.

    Args:
        inp (numpy.ndarray) : The input array.
        sigma (Union[float, List[float]]): Standard deviation for Gaussian kernel. The standard
            deviations of the Gaussian filter are given for each axis as a sequence, or as a
            single number, in which case it is equal for all axes.
        mode (str) : The mode parameter determines how the input array is extended when the
            filter overlaps a border. Default value is 'reflect'. The valid values and their
            behavior is as follows:

            * `constant`, (k k k k | a b c d | k k k k) : The input is extended by filling all
              values beyond the edge with the same constant value, defined by the `cval`
              parameter.
            * `nearest`, (a a a a | a b c d | d d d d) : The input is extended by replicating
              the last pixel.
            * `mirror`, (c d c b | a b c d | c b a b) : The input is extended by reflecting
              about the center of the last pixel. This mode is also sometimes referred to as
              whole-sample symmetric.
            * `reflect`, (d c b a | a b c d | d c b a) : The input is extended by reflecting
              about the edge of the last pixel. This mode is also sometimes referred to as
              half-sample symmetric.
            * `wrap`, (a b c d | a b c d | a b c d) : The input is extended by wrapping around
              to the opposite edge.

        cval (float) : Value to fill past edges of input if mode is ‘constant’. Default is 0.0.
        truncate (float) : Truncate the filter at this many standard deviations. Default is 4.0.
        backend (str) : Choose between numpy ('numpy') or FFTW ('fftw') backend library for the
            FFT implementation.
        num_threads (int) : Number of threads.

    Returns:
        numpy.ndarray : Gaussian gradient magnitude array. The array is the same shape as `input`.
    """
    cdef int ndim = inp.ndim
    cdef np.ndarray sigmas = normalize_sequence(sigma, ndim, np.NPY_FLOAT64)
    cdef double *_sig = <double *>np.PyArray_DATA(sigmas)

    cdef int n
    for n in range(ndim):
        if inp.shape[n] == 1:
            sigmas[n] = 0.0

    cdef int fail = 0
    cdef int _mode = extend_mode_to_code(mode)
    cdef np.npy_intp *dims = inp.shape
    cdef unsigned long *_dims = <unsigned long *>dims

    cdef np.ndarray out = <np.ndarray>np.PyArray_SimpleNew(ndim, dims, np.NPY_FLOAT64)
    cdef double *_out = <double *>np.PyArray_DATA(out)

    cdef void *_inp
    cdef int inp_tn = np.PyArray_TYPE(inp)
    if np.PyArray_ISCOMPLEX(inp):
        inp = check_array(inp, np.NPY_COMPLEX128)
        _inp = <double *>np.PyArray_DATA(inp)

        with nogil:
            if backend == 'fftw':
                fail = gauss_grad_mag_c(_out, <double complex *>_inp, ndim, _dims, _sig,
                                        _mode, <double complex>cval, truncate, num_threads,
                                        cfft_convolve_fftw)
            elif backend == 'numpy':
                fail = gauss_grad_mag_c(_out, <double complex *>_inp, ndim, _dims, _sig,
                                        _mode, <double complex>cval, truncate, num_threads,
                                        cfft_convolve_np)
            else:
                raise ValueError('{:s} is invalid backend'.format(backend))

    else:
        inp = check_array(inp, np.NPY_FLOAT64)
        _inp = <double *>np.PyArray_DATA(inp)

        with nogil:
            if backend == 'fftw':
                fail = gauss_grad_mag_r(_out, <double *>_inp, ndim, _dims, _sig, _mode,
                                        cval, truncate, num_threads, rfft_convolve_fftw)
            elif backend == 'numpy':
                fail = gauss_grad_mag_r(_out, <double *>_inp, ndim, _dims, _sig, _mode,
                                        cval, truncate, num_threads, rfft_convolve_np)
            else:
                raise ValueError('{:s} is invalid backend'.format(backend))
    
    if fail:
        raise RuntimeError('C backend exited with error.')
    return check_array(out, inp_tn)

def bar_positions(double x0, double x1, double b_dx, double rd, long seed):
    """Generate a coordinate array of randomized barcode's bar positions.

    Args:
        x0 (float) : Barcode's lower bound along the x axis [um].
        x1 (float) : Barcode's upper bound along the x axis [um].
        b_dx (float) : Average bar's size [um].
        rd (float) : Random deviation of barcode's bar positions (0.0 - 1.0).
        seed (int) : Seed used for pseudo random number generation.

    Returns:
        numpy.ndarray : Array of barcode's bar coordinates.
    """
    cdef np.npy_intp size = 2 * (<np.npy_intp>((x1 - x0) / 2 / b_dx) + 1) if x1 > x0 else 0
    cdef np.npy_intp *dims = [size,]
    cdef np.ndarray[double] bars = <np.ndarray>np.PyArray_SimpleNew(1, dims, np.NPY_FLOAT64)
    cdef double *_bars = <double *>np.PyArray_DATA(bars)
    if size:
        with nogil:
            barcode_bars(_bars, size, x0, b_dx, rd, seed)
    return bars

cdef np.ndarray ml_profile_wrapper(np.ndarray x_arr, np.ndarray layers, complex t0,
                                   complex t1, double sigma, unsigned num_threads):
    x_arr = check_array(x_arr, np.NPY_FLOAT64)
    layers = check_array(layers, np.NPY_FLOAT64)

    cdef int fail = 0
    cdef int ndim = x_arr.ndim
    cdef np.npy_intp *dims = x_arr.shape
    cdef np.ndarray out = <np.ndarray>np.PyArray_SimpleNew(ndim, dims, np.NPY_COMPLEX128)

    cdef np.npy_intp isize = np.PyArray_SIZE(x_arr)
    cdef np.npy_intp lsize = np.PyArray_SIZE(layers)
    cdef complex *_out = <complex *>np.PyArray_DATA(out)
    cdef double *_x = <double *>np.PyArray_DATA(x_arr)
    cdef double *_lyrs = <double *>np.PyArray_DATA(layers)
    with nogil:
        fail = ml_profile(_out, _x, isize, _lyrs, lsize, t0, t1, sigma, num_threads)
    if fail:
        raise RuntimeError('C backend exited with error.')
    return out

def barcode_profile(np.ndarray x_arr not None, np.ndarray bars not None, double bulk_atn,
                    double bar_atn, double bar_sigma, unsigned num_threads=1):
    r"""Return an array of barcode's transmission profile calculated
    at `x_arr` coordinates.

    Args:
        x_arr (numpy.ndarray) : Array of the coordinates, where the transmission
            coefficients are calculated [um].    
        bars (numpy.ndarray) : Coordinates of barcode's bar positions [um].
        bulk_atn (float) : Barcode's bulk attenuation coefficient (0.0 - 1.0).
        bar_atn (float) : Barcode's bar attenuation coefficient (0.0 - 1.0).
        bar_sigma (float) : Inter-diffusion length [um].
        num_threads (int) : Number of threads used in the calculations.
    
    Returns:
        numpy.ndarray : Array of barcode's transmission profiles.

    Notes:
        Barcode's transmission profile is given by:
        
        .. math::
            t_b(x) = t_{air} + \frac{t_1 - t_{air}}{2}
            \left( \tanh\left(\frac{x - x^b_1}{\sigma_b} \right) +
            \tanh\left(\frac{x^b_N - x}{\sigma_b}\right) \right)\\
            - \frac{t_2 - t_1}{2}\sum_{n = 1}^{(N - 1) / 2} 
            \left( \tanh\left(\frac{x - x^b_{2 n}}{\sigma_b}\right) + 
            \tanh\left(\frac{x^b_{2 n + 1} - x}{\sigma_b}\right) \right)
        
        where :math:`t_1 = \sqrt{1 - A_{bulk}}`, :math:`t_2 = \sqrt{1 - A_{bulk} - A_{bar}}`
        are the transmission coefficients for the bulk and the bars of the sample,
        :math:`x^b_n` is a set of bars coordinates, and :math:`\sigma_b` is the
        inter-diffusion length.
    """
    cdef complex t0 = sqrt(1.0 - bulk_atn)
    cdef complex t1 = sqrt(1.0 - bulk_atn - bar_atn)
    return ml_profile_wrapper(x_arr, bars, t0, t1, bar_sigma, num_threads)

def mll_profile(np.ndarray x_arr not None, np.ndarray layers not None, complex t0,
                complex t1, double sigma, unsigned num_threads=1):
    r"""Return an array of MLL's transmission profile calculated
    at `x_arr` coordinates.

    Args:
        x_arr (numpy.ndarray) : Array of the coordinates, where the transmission
            coefficients are calculated [um].    
        layers (numpy.ndarray) : Coordinates of MLL's layers positions [um].
        t0 (complex) : Fresnel transmission coefficient for the first material of
            MLL's bilayer.
        t1 (complex) : Fresnel transmission coefficient for the first material of
            MLL's bilayer.
        sigma (float) : Inter-diffusion length [um].
        num_threads (int) : Number of threads used in the calculations.
    
    Returns:
        numpy.ndarray : Array of barcode's transmission profiles.

    Notes:
        MLL's transmission profile is given by:
        
        .. math::
            t_{MLL}(x) = t_{air} + \frac{t_1 - t_{air}}{2}
            \left( \tanh\left(\frac{x - x^b_1}{\sigma_b} \right) +
            \tanh\left(\frac{x^b_N - x}{\sigma_b}\right) \right)\\
            + \frac{t_2 - t_1}{2}\sum_{n = 1}^{(N - 1) / 2} 
            \left( \tanh\left(\frac{x - x^b_{2 n}}{\sigma_b}\right) + 
            \tanh\left(\frac{x^b_{2 n + 1} - x}{\sigma_b}\right) \right)
        
        where :math:`t_1`, :math:`t_2` are the transmission coefficients of the
        MLL's bilayers, :math:`x^b_n` is a set of bilayer coordinates,
        and :math:`\sigma_b` is the inter-diffusion length.
    """
    return ml_profile_wrapper(x_arr, layers, t0, t1, sigma, num_threads)

def make_frames(np.ndarray pfx not None, np.ndarray pfy not None, double dx, double dy,
                tuple shape, long seed, unsigned num_threads=1):
    """Generate intensity frames from one-dimensional intensity profiles (`pfx`,
    `pfy`) and white-field profiles (`wfx`, `wfy`). Intensity profiles resized into
    the shape of a frame. Poisson noise is applied if `seed` is non-negative.

    Args:
        pfx (numpy.ndarray) : Intensity profile along the x axis.
        pfy (numpy.ndarray) : Intensity profile along the y axis.
        dx (float) : Sampling interval along the x axis [um].
        dy (float) : Sampling interval along the y axis [um].
        shape (Tuple[int, int]) : Shape of the detector array.
        seed (int) : Seed for pseudo-random number generation.
        num_threads (int) : Number of threads used in the calculations.

    Returns:
        numpy.ndarray : Intensity frames.
    """
    pfx = check_array(pfx, np.NPY_FLOAT64)
    pfy = check_array(pfy, np.NPY_FLOAT64)

    cdef int fail = 0
    cdef np.npy_intp *oshape = [pfx.shape[0], <np.npy_intp>(shape[0]), <np.npy_intp>(shape[1])]
    cdef np.ndarray out = <np.ndarray>np.PyArray_SimpleNew(3, oshape, np.NPY_FLOAT64)
    cdef unsigned long *_ishape = [<unsigned long>(pfx.shape[0]), <unsigned long>(pfy.shape[0]),
                                   <unsigned long>(pfx.shape[1])]
    cdef unsigned long *_oshape = [<unsigned long>(oshape[0]), <unsigned long>(oshape[1]), <unsigned long>(oshape[2])]
    cdef double *_out = <double *>np.PyArray_DATA(out)
    cdef double *_pfx = <double *>np.PyArray_DATA(pfx)
    cdef double *_pfy = <double *>np.PyArray_DATA(pfy)
    with nogil:
        fail = frames(_out, _pfx, _pfy, dx, dy, _ishape, _oshape, seed, num_threads)
    if fail:
        raise RuntimeError('C backend exited with error.')
    return out

def median(np.ndarray data not None, np.ndarray mask=None, int axis=0, unsigned num_threads=1):
    """Calculate a median along the `axis`.

    Args:
        data (numpy.ndarray) : Intensity frames.
        mask (numpy.ndarray) : Bad pixel mask.
        axis (int) : Array axis along which median values are calculated.
        num_threads (int) : Number of threads used in the calculations.

    Returns:
        numpy.ndarray : Array of medians along the given axis.
    """
    if not np.PyArray_IS_C_CONTIGUOUS(data):
        data = np.PyArray_GETCONTIGUOUS(data)

    cdef int ndim = data.ndim
    axis = axis if axis >= 0 else ndim + axis
    axis = axis if axis <= ndim - 1 else ndim - 1

    if mask is None:
        mask = <np.ndarray>np.PyArray_SimpleNew(ndim, data.shape, np.NPY_BOOL)
        np.PyArray_FILLWBYTE(mask, 1)
    else:
        mask = check_array(mask, np.NPY_BOOL)
        if memcmp(data.shape, mask.shape, ndim * sizeof(np.npy_intp)):
            raise ValueError('mask and data arrays must have identical shapes')

    cdef unsigned long *_dims = <unsigned long *>data.shape

    cdef np.npy_intp *odims = <np.npy_intp *>malloc((ndim - 1) * sizeof(np.npy_intp))
    if odims is NULL:
        raise MemoryError('not enough memory')
    cdef int i
    for i in range(axis):
        odims[i] = data.shape[i]
    for i in range(axis + 1, ndim):
        odims[i - 1] = data.shape[i]

    cdef int type_num = np.PyArray_TYPE(data)
    cdef np.ndarray out = <np.ndarray>np.PyArray_SimpleNew(ndim - 1, odims, type_num)
    cdef void *_out = <void *>np.PyArray_DATA(out)
    cdef void *_data = <void *>np.PyArray_DATA(data)
    cdef unsigned char *_mask = <unsigned char *>np.PyArray_DATA(mask)

    with nogil:
        if type_num == np.NPY_FLOAT64:
            fail = median_c(_out, _data, _mask, ndim, _dims, 8, axis, compare_double, num_threads)
        elif type_num == np.NPY_FLOAT32:
            fail = median_c(_out, _data, _mask, ndim, _dims, 4, axis, compare_float, num_threads)
        elif type_num == np.NPY_INT32:
            fail = median_c(_out, _data, _mask, ndim, _dims, 4, axis, compare_int, num_threads)
        elif type_num == np.NPY_UINT32:
            fail = median_c(_out, _data, _mask, ndim, _dims, 4, axis, compare_uint, num_threads)
        elif type_num == np.NPY_UINT64:
            fail = median_c(_out, _data, _mask, ndim, _dims, 8, axis, compare_ulong, num_threads)
        else:
            raise TypeError(f'data argument has incompatible type: {str(data.dtype)}')
    if fail:
        raise RuntimeError('C backend exited with error.')

    free(odims)
    return out

def median_filter(np.ndarray data not None, object size not None, np.ndarray mask=None,
                  str mode='reflect', double cval=0.0, unsigned num_threads=1):
    """Calculate a median along the `axis`.

    Args:
        data (numpy.ndarray) : Intensity frames.
        size (numpy.ndarray) : Gives the shape that is taken from the input array, at every
            element position, to define the input to the filter function. We adjust size to
            the number of dimensions of the input array, so that, if the input array is shape
            (10,10,10), and size is 2, then the actual size used is (2,2,2).
        mask (Optional[numpy.ndarray]) : Bad pixel mask.
        mode (str) : The mode parameter determines how the input array is extended when the
            filter overlaps a border. Default value is 'reflect'. The valid values and their
            behavior is as follows:

            * `constant`, (k k k k | a b c d | k k k k) : The input is extended by filling all
              values beyond the edge with the same constant value, defined by the `cval`
              parameter.
            * `nearest`, (a a a a | a b c d | d d d d) : The input is extended by replicating
              the last pixel.
            * `mirror`, (c d c b | a b c d | c b a b) : The input is extended by reflecting
              about the center of the last pixel. This mode is also sometimes referred to as
              whole-sample symmetric.
            * `reflect`, (d c b a | a b c d | d c b a) : The input is extended by reflecting
              about the edge of the last pixel. This mode is also sometimes referred to as
              half-sample symmetric.
            * `wrap`, (a b c d | a b c d | a b c d) : The input is extended by wrapping around
              to the opposite edge.
        cval (float) : Value to fill past edges of input if mode is 'constant'. Default is 0.0.
        num_threads (int) : Number of threads used in the calculations.

    Returns:
        numpy.ndarray : Filtered array. Has the same shape as `input`.
    """
    if not np.PyArray_IS_C_CONTIGUOUS(data):
        data = np.PyArray_GETCONTIGUOUS(data)

    cdef int ndim = data.ndim
    cdef np.npy_intp *dims = data.shape

    if mask is None:
        mask = <np.ndarray>np.PyArray_SimpleNew(ndim, dims, np.NPY_BOOL)
        np.PyArray_FILLWBYTE(mask, 1)
    else:
        mask = check_array(mask, np.NPY_BOOL)

    cdef np.ndarray fsize = normalize_sequence(size, ndim, np.NPY_INTP)
    cdef unsigned long *_fsize = <unsigned long *>np.PyArray_DATA(fsize)

    cdef unsigned long *_dims = <unsigned long *>dims
    cdef int type_num = np.PyArray_TYPE(data)
    cdef np.ndarray out = <np.ndarray>np.PyArray_SimpleNew(ndim, dims, type_num)
    cdef void *_out = <void *>np.PyArray_DATA(out)
    cdef void *_data = <void *>np.PyArray_DATA(data)
    cdef unsigned char *_mask = <unsigned char *>np.PyArray_DATA(mask)
    cdef int _mode = extend_mode_to_code(mode)
    cdef void *_cval = <void *>&cval

    with nogil:
        if type_num == np.NPY_FLOAT64:
            fail = median_filter_c(_out, _data, _mask, ndim, _dims, 8, _fsize, _mode, _cval, compare_double, num_threads)
        elif type_num == np.NPY_FLOAT32:
            fail = median_filter_c(_out, _data, _mask, ndim, _dims, 4, _fsize, _mode, _cval, compare_float, num_threads)
        elif type_num == np.NPY_INT32:
            fail = median_filter_c(_out, _data, _mask, ndim, _dims, 4, _fsize, _mode, _cval, compare_int, num_threads)
        elif type_num == np.NPY_UINT32:
            fail = median_filter_c(_out, _data, _mask, ndim, _dims, 4, _fsize, _mode, _cval, compare_uint, num_threads)
        elif type_num == np.NPY_UINT64:
            fail = median_filter_c(_out, _data, _mask, ndim, _dims, 8, _fsize, _mode, _cval, compare_ulong, num_threads)
        else:
            raise TypeError(f'data argument has incompatible type: {str(data.dtype)}')
    if fail:
        raise RuntimeError('C backend exited with error.')

    return out


""":class:`pyrost.simulation.STSim` does the heavy lifting of calculating the wavefront
propagation to the detector plane. :class:`pyrost.simulation.STConverter` exports
simulated data to a `CXI <https://www.cxidb.org/cxi.html>`_ format file accordingly to
the provided :class:`pyrost.CXIProtocol` object and saves the protocol and experimental
parameters to the same folder.

Examples:

    Perform the simulation for a given :class:`pyrost.simulation.STParams` object
    `params` with :class:`pyrost.simulation.STSim`. Then generate a stack frames with
    :func:`pyrost.simulation.STSim.frames`, and save the simulated data to a
    `CXI <https://www.cxidb.org/cxi.html>`_ file using the default protocol as follows:

    >>> import pyrost.simulation as st_sim
    >>> params = st_sim.STParams.import_default()
    >>> sim_obj = st_sim.STSim(params)
    >>> data = sim_obj.frames()
    >>> st_conv = st_sim.STConverter()
    >>> st_conv.save_sim(data, sim_obj, 'test') # doctest: +SKIP
"""
from __future__ import annotations
import os
import argparse
from typing import Iterable, Optional, Tuple, Union
import numpy as np
from ..cxi_protocol import CXIProtocol, CXIStore
from ..data_container import DataContainer, dict_to_object
from ..data_processing import STData, Crop
from .st_parameters import STParams
from ..bin import rsc_wp, fraunhofer_wp, fft_convolve
from ..bin import make_frames, gaussian_gradient_magnitude

class STSim(DataContainer):
    """One-dimensional Speckle Tracking scan simulation class.
    Generates barcode's transmission profile and lens' abberated
    wavefront, whereupon propagates the wavefront to the detector's
    plane.

    Attributes:
        attr_set : Set of attributes in the container which are necessary
            to initialize in the constructor.
        init_set : Set of optional data attributes.

    Notes:
        **Necessary attributes**:

        * backend : the backend library for the FFT implementation.
        * params : Experimental parameters.

        **Optional attributes**:

        * bars : Barcode's bar positions [um].
        * det_wfx : Wavefront at the detector plane along the x axis.
        * det_wfy : Wavefront at the detector plane along the y axis.
        * det_iy : Intensity profile at the detector plane along the
          x axis.
        * det_iy : Intensity profile at the detector plane along the
          y axis.
        * lens_wfx : Wavefront at the lens plane along the x axis.
        * lens_wfy : Wavefront at the lens plane along the y axis.
        * num_threads : Number of threads used in the calculations.
        * n_x : Wavefront size along the x axis.
        * n_y : Wavefront size along the y axis.
        * roi : Region of interest in detector plane.
        * smp_pos : Sample translations along the x axis.
        * smp_profile : Barcode's transmission profile.
        * smp_wfx : Wavefront at the sample plane along the x axis.
        * smp_wfy : Wavefront at the sample plane along the y axis.
    """
    backends = {'numpy', 'fftw'}
    attr_set = {'backend', 'params'}
    init_set = {'bars', 'det_wfx', 'det_wfy', 'det_ix', 'det_iy', 'lens_wfx',
                'lens_wfy', 'roi', 'smp_pos', 'smp_profile', 'smp_wfx', 'smp_wfy'}

    # Necessary attributes
    backend     : str
    params      : STParams

    # Automatically generated attributes
    bars        : np.ndarray
    det_ix      : np.ndarray
    det_iy      : np.ndarray
    det_wfx     : np.ndarray
    det_wfy     : np.ndarray
    lens_wfx    : np.ndarray
    lens_wfy    : np.ndarray
    roi         : Tuple[int, int, int, int]
    smp_pos     : np.ndarray
    smp_profile : np.ndarray
    smp_wfx     : np.ndarray
    smp_wfy     : np.ndarray

    def __init__(self, params: STParams, backend: str='numpy',
                 **kwargs: Union[int, np.ndarray]) -> None:
        """
        Args:
            params : Set of simulation parameters.
            backend : Choose between numpy ('numpy') or FFTW ('fftw') library for the FFT
                implementation.
            num_threads : Number of threads used in the calculations.
            kwargs : Any of the optional attributes specified in :class:`STSim` notes.

        Raises:
            ValueError : If the 'backend' keyword is invalid.
        """
        if not backend in self.backends:
            err_msg = f'backend must be one of the following: {str(self.backends):s}'
            raise ValueError(err_msg)

        super(STSim, self).__init__(backend=backend, params=params, **kwargs)

        self._init_functions(bars=lambda: self.params.bar_positions(dist=self.params.defocus),
                             lens_wfx=self.params.lens_x_wavefront, lens_wfy=self.params.lens_y_wavefront,
                             smp_wfx=self._sample_x_wavefront, smp_wfy=self._sample_y_wavefront,
                             smp_pos=self.params.sample_positions, smp_profile=self._sample_profile,
                             det_wfx=self._detector_x_wavefront, det_wfy=self._detector_y_wavefront,
                             det_ix=self._detector_x_intensity, det_iy=self._detector_y_intensity,
                             roi=self.find_beam_roi)

        self._init_attributes()

    def _sample_x_wavefront(self) -> np.ndarray:
        dx0 = 2.0 * self.params.ap_x / self.x_size
        dx1 = np.abs(dx0 * self.params.defocus / self.params.focus)
        z01 = self.params.focus + self.params.defocus
        return rsc_wp(wft=self.lens_wfx, dx0=dx0, dx=dx1, z=z01, wl=self.params.wl,
                        backend=self.backend, num_threads=self.params.num_threads)

    def _sample_y_wavefront(self) -> np.ndarray:
        dy0 = 2.0 * self.params.ap_y / self.y_size
        z01 = self.params.focus + self.params.defocus
        return rsc_wp(wft=self.lens_wfy, dx0=dy0, dx=dy0, z=z01, wl=self.params.wl,
                        backend=self.backend, num_threads=self.params.num_threads)


    def _sample_profile(self) -> np.ndarray:
        dx1 = np.abs(2.0 * self.params.ap_x * self.params.defocus / self.params.focus / self.x_size)
        x1_arr = dx1 * np.arange(-self.x_size // 2, self.x_size // 2) + self.smp_pos[:, None]
        return self.params.barcode_profile(x_arr=x1_arr, dx=dx1, bars=self.bars)

    def _detector_x_wavefront(self) -> np.ndarray:
        dx1 = np.abs(2.0 * self.params.ap_x * self.params.defocus / self.params.focus / self.x_size)
        dx2 = self.params.detx_size * self.params.pix_size / self.x_size
        wft = self.smp_wfx * self.smp_profile
        return fraunhofer_wp(wft=wft, dx0=dx1, dx=dx2, z=self.params.det_dist,
                             wl=self.params.wl, backend=self.backend,
                             num_threads=self.params.num_threads)

    def _detector_y_wavefront(self) -> np.ndarray:
        dy1 = 2.0 * self.params.ap_y / self.y_size
        dy2 = self.params.dety_size * self.params.pix_size / self.y_size
        return fraunhofer_wp(wft=self.smp_wfy, dx0=dy1, dx=dy2, z=self.params.det_dist,
                             wl=self.params.wl, backend=self.backend,
                             num_threads=1)

    def _detector_x_intensity(self) -> np.ndarray:
        dx = self.params.detx_size * self.params.pix_size / self.x_size
        sc_x = self.params.source_curve(dist=self.params.defocus + self.params.det_dist, step=dx)
        det_ix = np.sqrt(self.params.p0) / self.params.ap_x * np.abs(self.det_wfx)**2
        return fft_convolve(array=det_ix, kernel=sc_x, backend=self.backend,
                            num_threads=self.params.num_threads)

    def _detector_y_intensity(self) -> np.ndarray:
        dy = self.params.dety_size * self.params.pix_size / self.y_size
        sc_y = self.params.source_curve(dist=self.params.defocus + self.params.det_dist, step=dy)
        det_iy = np.sqrt(self.params.p0) / self.params.ap_y * np.abs(self.det_wfy)**2
        return fft_convolve(array=det_iy, kernel=sc_y, backend=self.backend,
                            num_threads=self.params.num_threads)

    def find_beam_roi(self) -> Tuple[int, int, int, int]:
        """Calculate the beam's field of view at the detector grid as a
        list of four pixel coordinates.

        Returns:
            A set of four coordinates ('y0', 'y1', 'x0', 'x1'), where 'y0' and
            'y1' are the lower and upper bounds of the beam along the vertical
            detector axis and 'x0' and 'x1' are the lower and upper bounds of
            the beam along the horizontal axis.
        """
        dx1 = np.abs(2.0 * self.params.ap_x * self.params.defocus / self.params.focus / self.x_size)
        dx2 = self.params.detx_size * self.params.pix_size / self.x_size
        wfield_x = np.abs(fraunhofer_wp(wft=self.smp_wfx, dx0=dx1, dx=dx2,
                                        z=self.params.det_dist, wl=self.params.wl,
                                        backend=self.backend, num_threads=1))

        dy1 = 2.0 * self.params.ap_y / self.y_size
        dy2 = self.params.dety_size * self.params.pix_size / self.y_size
        wfield_y = np.abs(fraunhofer_wp(wft=self.smp_wfy, dx0=dy1, dx=dy2,
                                        z=self.params.det_dist, wl=self.params.wl,
                                        backend=self.backend, num_threads=1))

        x0, x1 = self.params.beam_span(self.params.det_dist)

        if (x1 - x0) < self.params.detx_size * self.params.pix_size:
            dx = self.params.detx_size * self.params.pix_size / self.x_size
            cnt_x = self.x_size // 2 + int(0.5 * (x0 + x1) // dx)
            grad_x = gaussian_gradient_magnitude(wfield_x, self.x_size // 100, mode='nearest',
                                                 num_threads=self.params.num_threads)
            x0 = (np.argmax(grad_x[:cnt_x]) * self.params.detx_size) // self.x_size
            x1 = ((cnt_x + np.argmax(grad_x[cnt_x:])) * self.params.detx_size) // self.x_size
        else:
            x0, x1 = 0, self.params.detx_size

        cnt_y = self.y_size // 2
        grad_y = gaussian_gradient_magnitude(wfield_y, self.y_size // 100, mode='nearest',
                                             num_threads=self.params.num_threads)
        y0 = (np.argmax(grad_y[:cnt_y]) * self.params.dety_size) // self.y_size
        y1 = ((cnt_y + np.argmax(grad_y[cnt_y:])) * self.params.dety_size) // self.y_size
        return (y0, y1, x0, x1)

    @property
    def x_size(self) -> int:
        return self.lens_wfx.size

    @property
    def y_size(self) -> int:
        return self.lens_wfy.size

    @dict_to_object
    def update_bars(self, bars: np.ndarray) -> STSim:
        """Return a new :class:`STSim` object with the updated
        `bars`.

        Args:
            bars : Array of barcode's bar positions.

        Returns:
            A new :class:`STSim` object with the updated `bars`.
        """
        return {'bars': bars, 'smp_profile': None, 'det_wfx': None, 'det_ix': None}

    @dict_to_object
    def update_roi(self, roi: np.ndarray) -> STSim:
        """Return a new :class:`STSim` object with the updated
        region of interest.

        Args:
            roi : Region of interest in detector plane. The values are
            given in pixels as following : [`x0`, `x1`, `y0`, `y1`].

        Returns:
            A new :class:`STSim` object with the updated `roi`.
        """
        return {'roi': roi}

    def frames(self, wfieldx: Optional[np.ndarray]=None, wfieldy: Optional[np.ndarray]=None,
               apply_noise: bool=True) -> np.ndarray:
        """Return intensity frames at the detector plane. Applies
        Poisson noise if `apply_noise` is True.

        Args:
            wfieldx : Whitefield profile along the x axis.
            wfieldy : Whitefield profile along the y aixs.
            apply_noise : Apply Poisson noise if it's True.

        Returns:
            Intensity frames.
        """
        dx = self.params.detx_size * self.params.pix_size / self.x_size
        dy = self.params.dety_size * self.params.pix_size / self.y_size
        seed = self.params.seed if apply_noise else -1
        frames = make_frames(pfx=self.det_ix, pfy=self.det_iy, dx=dx, dy=dy,
                             shape=(self.params.dety_size, self.params.detx_size),
                             seed=seed, num_threads=self.params.num_threads)
        if not wfieldx is None:
            frames *= (wfieldx / wfieldx.mean())
        if not wfieldy is None:
            frames *= (wfieldy / wfieldy.mean())[:, None]
        return frames

    def ptychograph(self, wfieldx: Optional[np.ndarray]=None, wfieldy: Optional[np.ndarray]=None,
                    apply_noise: bool=True) -> np.ndarray:
        """Return a ptychograph of intensity frames. Applies Poisson
        noise if `apply_noise` is True.

        Args:
            wfieldx : Whitefield profile along the x axis.
            wfieldy : Whitefield profile along the y aixs.
            apply_noise : Apply Poisson noise if it's True.

        Returns:
            Ptychograph.
        """
        data = self.frames(wfieldx=wfieldx, wfieldy=wfieldy, apply_noise=apply_noise)
        return data.sum(axis=1)[:, None]

class STConverter(DataContainer):
    """
    Converter class to export simulated data from :class:`STSim` to a CXI
    file. Takes an instance of :class:`STSim` and a generated stack of frames
    and automatically generates all the attributes necessary for the robust
    speckle tracking reconstruction (see :class:`pyrost.STData`).

    Attributes:
        sim_obj : a :class:`STSim` instance.

    Notes:
        **Necessary attributes**:

        * data : Simulated intensity frames.

        **Automatically generated attributes**:

        * basis_vectors : Detector basis vectors.
        * defocus_x : Defocus distance along the horizontal detector axis.
        * defocus_y : Defocus distance along the vertical detector axis.
        * distance : Sample-to-detector distance.
        * translations : Sample's translations.
        * wavelength : Incoming beam's wavelength.
        * x_pixel_size : Pixel's size along the horizontal detector axis.
        * y_pixel_size : Pixel's size along the vertical detector axis.
    """
    unit_vector_fs = np.array([1, 0, 0])
    unit_vector_ss = np.array([0, -1, 0])
    attr_set = {'sim_obj', 'data', 'crd_rat'}
    init_set = {'basis_vectors', 'defocus_x', 'defocus_y', 'distance', 'translations',
                'wavelength', 'x_pixel_size', 'y_pixel_size'}

    # Necessary attributes
    sim_obj         : STSim
    data            : np.ndarray
    crd_rat         : float

    # Automatically generated attributes
    basis_vectors   : np.ndarray
    defocus_x       : float
    defocus_y       : float
    distance        : float
    translations    : np.ndarray
    wavelength      : float
    x_pixel_size    : float
    y_pixel_size    : float

    def __init__(self, sim_obj: STSim, data: np.ndarray, crd_rat: float=1e-6) -> None:
        """
        Args:
            sim_obj : a :class:`STSim` instance.
            data : Simulated stack of frames from `sim_obj`. The data may be
                generated either by :func:`STSim.frames` or :func:`STSim.ptychograph`
                method.
            crd_rat : Coordinates ratio between the simulated and saved data.
        """
        super(STConverter, self).__init__(sim_obj=sim_obj, data=data, crd_rat=crd_rat)

        self._init_functions(defocus_x=lambda: self.crd_rat * self.sim_obj.params.defocus,
                             defocus_y=lambda: self.crd_rat * self.sim_obj.params.defocus,
                             distance=lambda: self.crd_rat * self.sim_obj.params.det_dist,
                             wavelength=lambda: self.crd_rat * self.sim_obj.params.wl,
                             x_pixel_size=lambda: self.crd_rat * self.sim_obj.params.pix_size,
                             y_pixel_size=lambda: self.crd_rat * self.sim_obj.params.pix_size,
                             basis_vectors=self._basis_vectors, translations=self._translations)

        self._init_attributes()

    def _basis_vectors(self):
        pix_vec = np.array([self.y_pixel_size, self.x_pixel_size, 0])
        vec_fs = np.tile(pix_vec * self.unit_vector_fs, (self.sim_obj.params.n_frames, 1))
        vec_ss = np.tile(pix_vec * self.unit_vector_ss, (self.sim_obj.params.n_frames, 1))
        return np.stack((vec_ss, vec_fs), axis=1)

    def get_transform(self):
        """Return a :class:`pyrost.Crop` transform corresponding to the ROI of
        the beam on the detector.

        Returns:
            :class:`pyrost.Crop` transform corresponding to the ROI of the beam.
        """
        return Crop(self.sim_obj.roi)

    def _translations(self):
        t_arr = np.zeros((self.sim_obj.params.n_frames, 3))
        t_arr[:, 0] = -self.sim_obj.smp_pos
        return self.crd_rat * t_arr

    def export_data(self, out_path: str, apply_transform: bool=True,
                    protocol: CXIProtocol=CXIProtocol.import_default()) -> STData:
        """Save simulated data to the path specified in `out_path` and return
        :class:`pyrost.STData` object.

        Args:
            out_path : Path to the folder, where all the files are saved.
            apply_transform : Apply the cropping transform from :func:`STSim.get_transform`.
            protocol : CXI file protocol.

        Returns:
            Data container :class:`pyrost.STData`.
        """
        self.save(out_path, apply_transform, protocol)
        if apply_transform:
            return STData(CXIStore(out_path, protocol=protocol))
        return STData(CXIStore(out_path, protocol=protocol), transform=self.get_transform())

    def save(self, out_path: str, apply_transform: bool=True,
             protocol: CXIProtocol=CXIProtocol.import_default(),
             mode: str='overwrite', idxs: Optional[Iterable[int]]=None) -> None:
        """Save simulated data to the path specified in `out_path`.

        Args:
            out_path : Path to the folder, where all the files are saved.
            apply_transform :  Apply the cropping transform from :func:`STSim.get_transform`.
            protocol : CXI file protocol.
            mode : Writing mode:

                * `append` : Append the data array to already existing dataset.
                * `insert` : Insert the data under the given indices `idxs`.
                * `overwrite` : Overwrite the existing dataset.

            idxs : A set of frame indices where the data is saved if `mode` is
                `insert`.

        Returns:
            None
        """
        data_dict = {attr: self.get(attr) for attr in self.init_set}
        out_file = CXIStore(out_path, mode='a', protocol=protocol)
        data_obj = STData(input_file=CXIStore(out_path, protocol=protocol),
                          output_file=out_file, data=self.data, **data_dict)
        if apply_transform:
            data_obj = data_obj.update_transform(self.get_transform())
        data_obj.save(mode=mode, idxs=idxs)

def main():
    """Main fuction to run Speckle Tracking simulation and save the results to a CXI file.
    """
    parser = argparse.ArgumentParser(description="Run Speckle Tracking simulation",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('out_path', type=str, help="Output folder path")
    parser.add_argument('-f', '--ini_file', type=str,
                        help="Path to an INI file to fetch all of the simulation parameters")
    parser.add_argument('--defocus', type=float, help="Lens defocus distance, [um]")
    parser.add_argument('--det_dist', type=float,
                        help="Distance between the barcode and the detector [um]")
    parser.add_argument('--step_size', type=float, help="Scan step size [um]")
    parser.add_argument('--step_rnd', type=float,
                        help="Random deviation of sample translations [0.0 - 1.0]")
    parser.add_argument('--n_frames', type=int, help="Number of frames")
    parser.add_argument('--detx_size', type=int, help="horizontal axis frames size in pixels")
    parser.add_argument('--dety_size', type=int, help="vertical axis frames size in pixels")
    parser.add_argument('--p0', type=float, help="Source beam flux [cnt / s]")
    parser.add_argument('--wl', type=float, help="Wavelength [um]")
    parser.add_argument('--th_s', type=float, help="Source rocking curve width [rad]")
    parser.add_argument('--ap_x', type=float, help="Lens size along the x axis [um]")
    parser.add_argument('--ap_y', type=float, help="Lens size along the y axis [um]")
    parser.add_argument('--focus', type=float, help="Focal distance [um]")
    parser.add_argument('--alpha', type=float, help="Third order aberrations [rad/mrad^3]")
    parser.add_argument('--ab_cnt', type=float,
                        help="Lens' aberrations center point [0.0 - 1.0]")
    parser.add_argument('--bar_size', type=float, help="Average bar size [um]")
    parser.add_argument('--bar_sigma', type=float, help="Bar haziness width [um]")
    parser.add_argument('--bar_atn', type=float, help="Bar attenuation")
    parser.add_argument('--bulk_atn', type=float, help="Bulk attenuation")
    parser.add_argument('--bar_rnd', type=float, help="Bar random deviation")
    parser.add_argument('--offset', type=float,
                        help="Sample's offset at the beginning and the end of the scan [um]")
    parser.add_argument('-p', '--ptych', action='store_true', help="Generate ptychograph data")
    parser.set_defaults(**STParams.import_default())

    args = vars(parser.parse_args())
    out_path = args.pop('out_path')
    ini_file = args.pop('ini_file')
    is_ptych = args.pop('ptych')
    if ini_file:
        st_params = STParams.import_ini(ini_file)
    else:
        st_params = STParams.import_default(**args)

    sim_obj = STSim(st_params)
    if is_ptych:
        data = sim_obj.ptychograph()
    else:
        data = sim_obj.frames()
    STConverter(sim_obj, data).save(out_path, mode='overwrite')
    print(f"The simulation results have been saved to {os.path.abspath(out_path)}")
